import asyncio
import dataclasses as dc
import io
import textwrap as tw
import time
from typing import Sequence, Dict, Any, Optional, Tuple, Callable

import click
import networkx as nx
from asciidag.graph import Graph as AsciiDagGraph
from asciidag.node import Node as AsciiDagNode

import statey as st
from statey.executor import ExecutionInfo
from statey.plan import Plan, PlanNode
from statey.syms import utils, types, impl, Object
from statey.syms.path import PathParser
from statey.task import Task, SessionSwitch, ResourceGraphOperation, TaskStatus


def color_for_status(status: TaskStatus) -> Optional[str]:
    """
    Return the color (as a string) that we want to use to print this related to this status
    """
    color_dict = {
        TaskStatus.SUCCESS: "green",
        TaskStatus.FAILED: "red",
        TaskStatus.SKIPPED: "yellow",
    }
    return color_dict.get(status)


def truncate_string(value: str, length: int) -> str:
    unstyled_len = len(click.unstyle(value))
    styled_len = len(value)

    style_chars = styled_len - unstyled_len

    if unstyled_len <= length:
        return value
    return value[: length - style_chars] + "..."


def ljust(value: str, length: int) -> str:
    unstyled_len = len(click.unstyle(value))
    styled_len = len(value)
    style_chars = styled_len - unstyled_len

    return value.ljust(length + style_chars)


class ColoredTypeRenderer(types.TypeStringRenderer):
    """
    Type string renderer that will add coloration
    """

    style_map = {
        types.TypeStringToken.RIGHT_BRACE: lambda x: click.style(x, fg="yellow"),
        types.TypeStringToken.LEFT_BRACE: lambda x: click.style(x, fg="yellow"),
        types.TypeStringToken.COLON: lambda x: click.style(x, fg="yellow"),
        types.TypeStringToken.COMMA: lambda x: click.style(x, fg="yellow"),
        types.TypeStringToken.TYPE_NAME: lambda x: click.style(x, fg="cyan"),
    }
    state_style_map = {
        "UP": lambda x: click.style(x, fg="green", bold=True),
        "DOWN": lambda x: click.style(x, fg="red", bold=True),
        None: lambda x: click.style(x, fg="yellow", bold=True),
    }

    def render(self, value: str, token: types.TypeStringToken) -> str:
        if token in self.style_map:
            return self.style_map[token](value)
        return value

    def render_state(self, name: str) -> str:
        return self.state_style_map.get(name, self.state_style_map[None])(name)


def data_to_lines(data: Any, name_func=lambda x: x) -> Sequence[str]:
    """
    Render the data to a readable, yaml-like structure
    """
    if isinstance(data, Object) and isinstance(data._impl, impl.Unknown):
        return ["<unknown>"]
    if isinstance(data, list):
        out_lines = []
        for item in data:
            lines = [
                tw.indent(line, "- " if idx == 0 else "  ")
                for idx, line in enumerate(data_to_lines(item, name_func=name_func))
            ]
            out_lines.extend(lines)
        return out_lines
    if isinstance(data, dict):
        out_lines = []
        for key in sorted(data):
            val = data[key]
            lines = data_to_lines(val, name_func=name_func)
            if len(lines) <= 1:
                out_lines.append(f'{name_func(key)}: {"".join(lines)}')
            else:
                out_lines.append(f"{name_func(key)}:")
                out_lines.extend(tw.indent(line, "  ") for line in lines)
        return out_lines

    return [repr(data)]


def data_to_string(data: Any, name_func=lambda x: x) -> str:
    """
    Give a nice human-readable representaiton of the data using YAML
    """
    return "\n".join(data_to_lines(data, name_func=name_func))


@dc.dataclass(frozen=True)
class PlanNodeSummary:
    """
    Contains basic human-redable details about a plan node
    """

    show_tasks: nx.DiGraph
    plan_node: PlanNode

    def _style_config_name(self, name: str) -> str:
        return click.style(name, bold=True)

    def _style_current_name(self, name: str) -> str:
        return click.style(name, bold=True)

    def crud_figures(
        self, task_graph: nx.DiGraph, show_metatasks: bool = False
    ) -> Tuple[int, int, int]:
        """
        Returns (creating, updating, deleting) tuple
        """
        creating, updating, deleting = 0, 0, 0

        if self.plan_node.current_action:
            if self.plan_node.current_state:
                deleting += 1

            if self.plan_node.config_action and self.plan_node.config_state:
                creating += 1
        elif self.plan_node.current_state:
            for node in task_graph.nodes:
                if node.startswith(self.plan_node.key + ":"):
                    task = task_graph.nodes[node]["task"]
                    if show_metatasks or not task.is_metatask():
                        updating += 1
                        break
        else:
            creating += 1

        return (creating, updating, deleting)

    def _config_summary(self) -> str:
        if (
            not self.plan_node.config_type
            or self.plan_node.config_type == types.EmptyType
        ):
            return "<none>"
        return data_to_string(
            self.plan_node.config_data, name_func=self._style_config_name
        )

    def _current_summary(self) -> str:
        if (
            not self.plan_node.current_type
            or self.plan_node.current_type == types.EmptyType
        ):
            return "<none>"
        return data_to_string(
            self.plan_node.current_data, name_func=self._style_current_name
        )

    def data_to_string(self, max_width: int) -> str:

        should_diff = False
        if (
            self.plan_node.current_type == self.plan_node.config_type
            and self.plan_node.current_type != types.EmptyType
        ):
            should_diff = True

        if (
            not should_diff
            and self.plan_node.current_type
            and self.plan_node.config_type
        ):
            try:
                caster = st.registry.get_caster(
                    self.plan_node.current_type, self.plan_node.config_type
                )
            except st.exc.NoCasterFound:
                pass
            else:
                from statey.syms.casters import ForceCaster

                if isinstance(caster, ForceCaster):
                    should_diff = True

        if should_diff:
            differ = st.registry.get_differ(self.plan_node.current_type)
            config_differ = None
            if self.plan_node.config_state:
                config_differ = st.registry.get_differ(
                    self.plan_node.config_state.input_type
                )

            diff = differ.diff(self.plan_node.current_data, self.plan_node.config_data)
            config_diff = []
            if config_differ is not None:
                config_diff = config_differ.diff(
                    self.plan_node.current_data, self.plan_node.config_data
                )

            current_lines = []
            config_lines = []
            path_parser = PathParser()

            if not diff:
                return click.style("<no diff>", bold=True)

            def diff_style_current_name(name):
                og_styled = self._style_current_name(name)
                if name in config_diff:
                    return click.style(og_styled, bg="red")
                return og_styled

            def diff_style_config_name(name):
                og_styled = self._style_config_name(name)
                if name in config_diff:
                    return click.style(og_styled, bg="red")
                return og_styled

            for subdiff in diff:
                current_diff_lines = data_to_lines(
                    subdiff.left, name_func=diff_style_current_name
                )
                path_str = path_parser.join(subdiff.path)

                if len(current_diff_lines) <= 1:
                    current_lines.append(
                        f"{diff_style_current_name(path_str)}:"
                        f' {"".join(current_diff_lines)}'
                    )
                else:
                    current_lines.append(f"{diff_style_current_name(path_str)}:")
                    current_lines.extend(
                        map(lambda x: tw.indent(x, "  "), current_diff_lines)
                    )

                config_diff_lines = data_to_lines(
                    subdiff.right, name_func=diff_style_config_name
                )
                if len(config_diff_lines) <= 1:
                    config_lines.append(
                        f"{diff_style_config_name(path_str)}:"
                        f' {"".join(config_diff_lines)}'
                    )
                else:
                    config_lines.append(f"{diff_style_config_name(path_str)}:")
                    config_lines.extend(
                        map(lambda x: tw.indent(x, "  "), config_diff_lines)
                    )

                if len(config_lines) > len(current_lines):
                    current_lines.extend(
                        [click.style("", bold=True)]
                        * (len(config_lines) - len(current_lines))
                    )

                if len(current_lines) > len(config_lines):
                    config_lines.extend(
                        [click.style("", bold=True)]
                        * (len(current_lines) - len(config_lines))
                    )

        else:
            current_lines = self._current_summary().split("\n")
            config_lines = self._config_summary().split("\n")

        max_n_lines = max(len(current_lines), len(config_lines))

        for lines in [current_lines, config_lines]:
            while len(lines) < max_n_lines:
                lines.append("")

        if not current_lines:
            return ""

        max_line_length = max(map(lambda x: len(click.unstyle(x)), current_lines))
        buffer_length = 6

        column_split = min((max_width - buffer_length) // 2, max_line_length)

        current_lines = [truncate_string(line, column_split) for line in current_lines]
        config_lines = [
            truncate_string(line, max_width - column_split - buffer_length)
            for line in config_lines
        ]

        def sep(idx):
            return click.style("  =>  ", fg="yellow", bold=True)
            # return ' => ' if idx % 2 == 0 else '    '

        out_lines = [
            "".join([ljust(current_line, column_split), sep(idx), config_line])
            for idx, (current_line, config_line) in enumerate(
                zip(current_lines, config_lines)
            )
        ]
        return "\n".join(out_lines)

    def to_string(self, max_width: int, indent: int = 2) -> str:
        if all(
            self.show_tasks.nodes[node]["task"].is_metatask()
            for node in self.show_tasks.nodes
        ):
            return ""

        style_key = lambda x: click.style(x, fg="green", bold=True)

        indent_str = " " * indent
        data_string = self.data_to_string(max_width)

        tasks_lines = [f'- {style_key("%s task(s):" % len(self.show_tasks.nodes))}']
        for node in nx.topological_sort(self.show_tasks):
            task = self.show_tasks.nodes[node]["task"]
            desc_str = ": " + task.description if task.description else ""
            line = f'- {click.style(node, fg="yellow")}{desc_str}'
            line = tw.indent(line, indent_str)
            tasks_lines.append(line)

        tasks_string = "\n".join(tasks_lines)

        type_lines = []
        renderer = ColoredTypeRenderer()
        if (
            self.plan_node.current_type is not None
            and self.plan_node.current_type != self.plan_node.config_type
        ):
            rendered = self.plan_node.current_type.render_type_string(renderer)
            type_string = f'- {style_key("type (current)")}: {rendered}'
            type_lines.append(type_string)

        state_lines = []
        if self.plan_node.current_state is not None and (
            self.plan_node.config_state is None
            or self.plan_node.current_state.state.name
            != self.plan_node.config_state.state.name
        ):
            state = renderer.render_state(self.plan_node.current_state.state.name)
            state_lines.append(f'- {style_key("state (current)")}: {state}')

        if self.plan_node.config_type is not None:
            rendered = self.plan_node.config_type.render_type_string(renderer)
            type_string = f'- {style_key("type")}: {rendered}'
            type_lines.append(type_string)

        if self.plan_node.config_state is not None:
            state = renderer.render_state(self.plan_node.config_state.state.name)
            state_lines.append(f'- {style_key("state")}: {state}')

        type_string = "\n".join(type_lines)
        state_string = "\n".join(state_lines)

        data_string_with_title = "\n".join(
            [f'- {style_key("data")}:', tw.indent(data_string, indent_str * 2)]
        )

        resource = (
            self.plan_node.config_state.resource
            if self.plan_node.config_state
            else self.plan_node.current_state.resource
        )

        data_string = tw.indent(data_string, indent_str * 2)
        lines = [
            f'- {style_key(self.plan_node.key)}{click.style("[", fg="yellow")}'
            f'{click.style(resource, fg="cyan")}{click.style("]", fg="yellow")}:',
            tw.indent(tasks_string, indent_str),
            tw.indent(type_string, indent_str),
            tw.indent(state_string, indent_str),
            tw.indent(data_string_with_title, indent_str),
        ]
        return "\n".join(lines)


def ascii_dag(graph: nx.DiGraph, key: Callable[[str], str] = lambda x: x) -> str:
    """
    Render the given DAG as a string, showing dependencies
    """
    outfile = io.StringIO()
    ascii_graph = AsciiDagGraph(outfile)

    nodes = {}
    tips = []

    for node in reversed(list(nx.topological_sort(graph))):
        parent_nodes = [nodes[path] for path in graph.succ[node]]
        ascii_node = nodes[node] = AsciiDagNode(key(node), parents=parent_nodes)
        if not graph.pred[node]:
            tips.append(ascii_node)

    ascii_graph.show_nodes(tips)
    return outfile.getvalue()


@dc.dataclass(frozen=True)
class PlanSummary:
    """
    Contains basic human-readable details about a plan
    """

    plan: Plan
    node_summaries: Sequence[PlanNodeSummary]
    show_metatasks: bool

    def non_empty_summaries(
        self, max_width: int, indent: int = 2
    ) -> Sequence[PlanNodeSummary]:
        summaries = []
        for node in self.node_summaries:
            summary = node.to_string(max_width, indent)
            if summary:
                summaries.append(summary)
        return summaries

    def to_string(self, max_width: int, indent: int = 2) -> str:
        summaries = self.non_empty_summaries(max_width, indent)
        return "\n\n".join(summaries)

    def task_dag_string(self) -> str:
        task_graph = self.plan.task_graph.task_graph.copy()
        keep = set()

        for node in task_graph:
            task = task_graph.nodes[node]["task"]
            if self.show_metatasks or not task.is_metatask():
                keep.add(node)

        utils.subgraph_retaining_dependencies(task_graph, keep)
        return ascii_dag(graph=task_graph, key=lambda x: click.style(x, fg="yellow"))

    def short_summary_string(self) -> str:
        creating, updating, deleting = 0, 0, 0

        for summary in self.node_summaries:
            new_creating, new_updating, new_deleting = summary.crud_figures(
                self.plan.task_graph.task_graph
            )
            creating += new_creating
            updating += new_updating
            deleting += new_deleting

        return (
            f'{creating} to {click.style("create", fg="green", bold=True)}, '
            f'{updating} to {click.style("update", fg="yellow", bold=True)}, '
            f'{deleting} to {click.style("delete", fg="red", bold=True)}.'
        )


@dc.dataclass(frozen=True)
class ExecutionSummary:
    """
    Summary info about executioninfo
    """

    exec_info: ExecutionInfo
    show_metatasks: bool

    def tasks_by_status(self) -> Dict[TaskStatus, Sequence[str]]:
        """
        Return all task keys in this task graph grouped by their status
        """
        out = {}
        for node in self.exec_info.task_graph.task_graph.nodes:
            status = self.exec_info.task_graph.get_info(node).status
            task = self.exec_info.task_graph.get_task(node)
            if (
                not self.show_metatasks
                and task.is_metatask()
                and status == TaskStatus.SUCCESS
            ):
                continue
            out.setdefault(status, []).append(node)
        return out

    def to_string(self, indent: int = 2, full_trace: bool = False) -> str:
        """
        Render a human-readable view describing what occurred in this
        exec info
        """
        task_dict = self.tasks_by_status()
        duration = (
            self.exec_info.end_timestamp - self.exec_info.start_timestamp
        ).total_seconds()

        run_time_line = click.style(f"Run time: {duration:.2f}s.", bold=True)

        failed_tasks = task_dict.get(TaskStatus.FAILED, [])
        skipped_tasks = task_dict.get(TaskStatus.SKIPPED, [])
        success_tasks = task_dict.get(TaskStatus.SUCCESS, [])

        if task_dict:
            task_lines = [f"Tasks completed by status:"]
            for status in sorted(task_dict, key=lambda x: x.value):
                status_name = click.style(
                    status.name, bold=True, fg=color_for_status(status)
                )
                task_lines.append(f"- {status_name}: {len(task_dict[status])} task(s)")
            task_line = "\n".join(task_lines)
        else:
            task_line = click.style("No tasks executed", bold=True, fg="yellow")

        is_success = (
            not self.exec_info.cancelled_by_errors and not self.exec_info.cancelled_by
        )
        if is_success:
            lines = [
                click.style("Plan executed successfully.", fg="green", bold=True),
                task_line,
                "",
                run_time_line,
            ]
            return "\n".join(lines)

        if failed_tasks:
            lines = [
                click.style(f"Plan execution failed!", fg="red", bold=True),
                task_line,
                "",
                run_time_line,
            ]
            for task_name in failed_tasks:
                error = self.exec_info.task_graph.get_info(task_name).error
                name = click.style(
                    task_name, bold=True, fg=color_for_status(TaskStatus.FAILED)
                )
                if full_trace:
                    lines.append(f"- {name}:\n{error.format_exception()}\n")
                else:
                    lines.append(f"- {name}: {error.format_error_message()}")
            return "\n".join(lines)

        lines = [
            click.style(
                f"Statey exited unexpectedly during execution!", fg="yellow", bold=True
            ),
            task_line,
            "",
            run_time_line,
        ]
        return "\n".join(lines)


@dc.dataclass(frozen=True)
class Inspector:
    """
    Utilities for visualizing plans in a human-readable manner.
    """

    def plan_summary(self, plan: Plan, show_metatasks: bool = False) -> PlanSummary:
        nodes = []

        for node in plan.nodes:
            task_graph_nodes = [
                task_node
                for task_node in plan.task_graph.task_graph.nodes
                if task_node.startswith(f"{node.key}:")
            ]

            remove = set()
            for sub_node in task_graph_nodes:
                task = plan.task_graph.get_task(sub_node)
                if not show_metatasks and task.is_metatask():
                    remove.add(sub_node)

            keep = set(task_graph_nodes) - remove
            task_graph = plan.task_graph.task_graph.copy()
            utils.subgraph_retaining_dependencies(task_graph, keep)

            summary = PlanNodeSummary(task_graph, node)
            nodes.append(summary)

        return PlanSummary(plan, nodes, show_metatasks)

    def execution_summary(
        self, exec_info: ExecutionInfo, show_metatasks: bool = False
    ) -> ExecutionSummary:
        """
        Get a summary for an exec_info instance
        """
        return ExecutionSummary(exec_info, show_metatasks)


def simple_print_graph(graph: nx.DiGraph, print_func=print) -> None:
    print_func("Nodes:")
    for node in graph.nodes:
        print_func(f"{node}: {graph.nodes[node]}")

    print_func()
    print_func("Edges:")
    for tup in graph.edges:
        if len(tup) == 3:
            left, right, idx = tup
            print_func(f"{left} -> {right}[{idx}]: {graph.edges[left, right, idx]}")
        else:
            left, right = tup
            print_func(f"{left} -> {right}: {graph.edges[left, right]}")



import abc
import dataclasses as dc
from typing import Any, Mapping, Optional, Dict

import marshmallow as ma

import statey as st
from statey.syms import (
    impl,
    impl as impl_mod,
    Object,
    session,
    type_serializers as ts,
    utils,
    encoders,
    semantics,
    types,
)


class ObjectImplementationSerializer(abc.ABC):
    """
    Serializes an ObjectImplementation
    """

    @abc.abstractmethod
    def serialize(self, implementation: impl.ObjectImplementation) -> Any:
        """
        Serialize the given object implementation into a JSON-serializable
        format
        """
        raise NotImplementedError

    @abc.abstractmethod
    def deserialize(
        self, data: Any, session: session.Session, objects: Mapping[Any, Object]
    ) -> impl.ObjectImplementation:
        """
        Given an object decoded from JSON, construct an ObjectImplementation
        instance
        """
        raise NotImplementedError


class ReferenceSchema(ma.Schema):
    """
    Schema for serializing Reference data
    """

    name = ma.fields.Str(required=True, default="reference")
    path = ma.fields.Str(required=True)


class ReferenceSerializer(ObjectImplementationSerializer):
    """
    Serializes a Reference
    """

    def serialize(self, implementation: impl.ObjectImplementation) -> Any:
        out = {"path": implementation.path}
        return ReferenceSchema().dump(out)

    def deserialize(
        self, data: Any, session: session.Session, objects: Mapping[Any, Object]
    ) -> impl.ObjectImplementation:
        deserialized = ReferenceSchema().load(data)
        return impl.Reference(deserialized["path"], session.ns)

    @classmethod
    @st.hookimpl
    def get_impl_serializer(
        cls, impl: impl.ObjectImplementation, type: types.Type, registry: "Registry"
    ) -> ObjectImplementationSerializer:
        if not isinstance(impl, impl_mod.Reference):
            return None
        return cls()

    @classmethod
    @st.hookimpl
    def get_impl_serializer_from_data(
        cls, data: Any, registry: "Registry"
    ) -> ObjectImplementationSerializer:
        if not isinstance(data, dict) or data.get("name") != "reference":
            return None
        return cls()


class FutureSchema(ma.Schema):
    """
    Schema for serializing Future data
    """

    name = ma.fields.Str(required=True, default="future")
    refs = ma.fields.List(ma.fields.Field(), required=True, default=list)
    return_type = ma.fields.Dict(missing=None, default=None)
    result = ma.fields.Field(missing=utils.MISSING)


@dc.dataclass(frozen=True)
class FutureSerializer(ObjectImplementationSerializer):
    """
    Serializes a Future
    """

    return_type_serializer: Optional[ts.TypeSerializer] = None

    def serialize(self, implementation: impl.ObjectImplementation) -> Any:
        out = {"refs": [obj._impl.id for obj in implementation.refs]}
        if (
            self.return_type_serializer is not None
            and implementation.return_type is not None
        ):
            out["return_type"] = self.return_type_serializer.serialize(
                implementation.return_type
            )

        if not utils.is_missing(implementation.result.result):
            out["result"] = implementation.result.result

        return FutureSchema().dump(out)

    def deserialize(
        self, data: Any, session: session.Session, objects: Mapping[Any, Object]
    ) -> impl.ObjectImplementation:
        deserialized = FutureSchema().load(data)
        result_value = data["result"]
        result = impl.FutureResult(result_value)

        object_refs = [objects[ref_id] for ref_id in data["refs"]]

        return_type = None
        if data["return_type"] is not None and self.return_type_serializer is not None:
            return_type = self.return_type_serializer.deserialize(data["return_type"])

        return impl.Future(object_refs, result, return_type)

    @classmethod
    @st.hookimpl
    def get_impl_serializer(
        cls, impl: impl.ObjectImplementation, type: types.Type, registry: "Registry"
    ) -> ObjectImplementationSerializer:
        if not isinstance(impl, impl_mod.Future):
            return None
        return_type_serializer = None
        if impl.return_type is not None:
            return_type_serializer = registry.get_type_serializer(impl.return_type)
        return cls(return_type_serializer)

    @classmethod
    @st.hookimpl
    def get_impl_serializer_from_data(
        cls, data: Any, registry: "Registry"
    ) -> ObjectImplementationSerializer:
        if not isinstance(data, dict) or data.get("name") != "future":
            return None
        return_type_serializer = None
        if data["return_type"] is not None:
            return_type_serializer = registry.get_type_serializer_from_data(
                data["return_type"]
            )
        return cls(return_type_serializer)


class UnknownSchema(ma.Schema):
    """
    Schema for serializing Unknown data
    """

    name = ma.fields.Str(required=True, default="unknown")
    obj = ma.fields.Field()
    refs = ma.fields.List(ma.fields.Field(), required=True, default=list)
    return_type = ma.fields.Dict(missing=None, default=None)


@dc.dataclass(frozen=True)
class UnknownSerializer(ObjectImplementationSerializer):
    """
    Serializes an Unknown value
    """

    return_type_serializer: Optional[ts.TypeSerializer] = None

    def serialize(self, implementation: impl.ObjectImplementation) -> Any:
        out = {"refs": [obj._impl.id for obj in implementation.refs]}
        if implementation.obj is not None:
            out["obj"] = implementation.obj._impl.id

        if (
            implementation.return_type is not None
            and self.return_type_serializer is not None
        ):
            out["return_type"] = self.return_type_serializer.serialize(
                implementation.return_type
            )

        return UnknownSchema().dump(out)

    def deserialize(
        self, data: Any, session: session.Session, objects: Mapping[Any, Object]
    ) -> impl.ObjectImplementation:
        deserialized = UnknownSchema().load(data)
        refs = [objects[ref_id] for ref_id in deserialized["refs"]]

        obj = deserialized["obj"]
        if obj is not None:
            obj = objects[obj]

        return_type = None
        if (
            deserialized["return_type"] is not None
            and self.return_type_serializer is not None
        ):
            return_type = self.return_type_serializer.deserialize(
                deserialized["return_type"]
            )

        return impl.Unknown(obj, refs, return_type)

    @classmethod
    @st.hookimpl
    def get_impl_serializer(
        cls, impl: impl.ObjectImplementation, type: types.Type, registry: "Registry"
    ) -> ObjectImplementationSerializer:
        if not isinstance(impl, impl_mod.Unknown):
            return None
        return_type_serializer = registry.get_type_serializer(type)
        return cls(return_type_serializer)

    @classmethod
    @st.hookimpl
    def get_impl_serializer_from_data(
        cls, data: Any, registry: "Registry"
    ) -> ObjectImplementationSerializer:
        if not isinstance(data, dict) or data.get("name") != "unknown":
            return None
        return_type_serializer = None
        if data["return_type"] is not None:
            return_type_serializer = registry.get_type_serializer_from_data(
                data["return_type"]
            )
        return cls(return_type_serializer)


class StructFieldSchema(ma.Schema):
    """
    Schema for a StructField
    """

    name = ma.fields.Str(required=True)
    value = ma.fields.Field(required=True)


class StructSchema(ma.Schema):
    """
    Schema for a Struct
    """

    name = ma.fields.Str(required=True, default="struct")
    fields = ma.fields.List(
        ma.fields.Nested(StructFieldSchema()), required=True, default=list
    )


@dc.dataclass(frozen=True)
class StructSerializer(ObjectImplementationSerializer):
    """
    Serializes a Struct value
    """

    def serialize(self, implementation: impl.ObjectImplementation) -> Any:
        fields = []
        field_schema = StructFieldSchema()
        for field in implementation.fields:
            fields.append({"name": field.name, "value": field.value._impl.id})

        out = {"fields": fields}
        return StructSchema().dump(out)

    def deserialize(
        self, data: Any, session: session.Session, objects: Mapping[Any, Object]
    ) -> impl.ObjectImplementation:
        deserialized = StructSchema().load(data)
        fields = []
        for field in deserialized["fields"]:
            name = field["name"]
            value = objects[field["value"]]
            fields.append(impl.StructField(name, value))
        return impl.Struct(fields)

    @classmethod
    @st.hookimpl
    def get_impl_serializer(
        cls, impl: impl.ObjectImplementation, type: types.Type, registry: "Registry"
    ) -> ObjectImplementationSerializer:
        if not isinstance(impl, impl_mod.Struct):
            return None
        return cls()

    @classmethod
    @st.hookimpl
    def get_impl_serializer_from_data(
        cls, data: Any, registry: "Registry"
    ) -> ObjectImplementationSerializer:
        if not isinstance(data, dict) or data.get("name") != "struct":
            return None
        return cls()


class ExpectedValueSchema(ma.Schema):
    """
    Schema for an ExpectedValue
    """

    name = ma.fields.Str(required=True, default="expected_value")
    obj = ma.fields.Field(required=True)
    expected = ma.fields.Field(required=True)


@dc.dataclass(frozen=True)
class ExpectedValueSerializer(ObjectImplementationSerializer):
    """
    Serializes an ExpectedValue value
    """

    def serialize(self, implementation: impl.ObjectImplementation) -> Any:
        out = {
            "obj": implementation.obj._impl.id,
            "expected": implementation.expected._impl.id,
        }
        return ExpectedValueSchema().dump(out)

    def deserialize(
        self, data: Any, session: session.Session, objects: Mapping[Any, Object]
    ) -> impl.ObjectImplementation:
        deserialized = ExpectedValueSchema().load(data)
        obj = objects[deserialized["obj"]]
        expected = objects[deserialized["expected"]]
        return impl.ExpectedValue(obj, expected)

    @classmethod
    @st.hookimpl
    def get_impl_serializer(
        cls, impl: impl.ObjectImplementation, type: types.Type, registry: "Registry"
    ) -> ObjectImplementationSerializer:
        if not isinstance(impl, impl_mod.ExpectedValue):
            return None
        return cls()

    @classmethod
    @st.hookimpl
    def get_impl_serializer_from_data(
        cls, data: Any, registry: "Registry"
    ) -> ObjectImplementationSerializer:
        if not isinstance(data, dict) or data.get("name") != "expected_value":
            return None
        return cls()


class DataSerializer(ObjectImplementationSerializer):
    """
    Base class for serializers for Data implementations
    """


class DataSchema(ma.Schema):
    """
    Schema for a data value w/o an algebraic type
    """

    name = ma.fields.Str(required=True, default="data")
    value = ma.fields.Field(required=True)
    type = ma.fields.Dict(required=True)
    value_type = ma.fields.Dict(missing=None, default=None)


@dc.dataclass(frozen=True)
class DataValueSerializer(DataSerializer):
    """
    Serializes a Data value w/ a non-algebraic data type
    """

    value_encoder: encoders.Encoder
    type_serializer: ts.TypeSerializer
    value_type_serializer: Optional[ts.TypeSerializer] = None

    def serialize(self, implementation: impl.ObjectImplementation) -> Any:
        out = {"type": self.type_serializer.serialize(self.value_encoder.type)}
        if isinstance(implementation.value, Object):
            out["value"] = {"object": implementation.value._impl.id}
        else:
            out["value"] = {"data": self.value_encoder.encode(implementation.value)}

        if (
            implementation.value_type is not None
            and self.value_type_serializer is not None
        ):
            out["value_type"] = self.value_type_serializer.serialize(
                implementation.value_type
            )

        return DataSchema().dump(out)

    def deserialize(
        self, data: Any, session: session.Session, objects: Mapping[Any, Object]
    ) -> impl.ObjectImplementation:

        deserialized = DataSchema().load(data)
        val_json = deserialized["value"]
        if "object" in val_json:
            value = objects[val_json["object"]]
        elif "data" in val_json:
            value = self.value_encoder.decode(val_json["data"])
        else:
            raise ValueError(f"Unable to interpret {val_json}!")

        value_type = None
        if (
            deserialized["value_type"] is not None
            and self.value_type_serializer is not None
        ):
            value_type = self.value_type_serializer.deserialize(
                deserialized["value_type"]
            )

        return impl.Data(value, value_type)

    @classmethod
    @st.hookimpl
    def get_impl_serializer(
        cls, impl: impl.ObjectImplementation, type: types.Type, registry: "Registry"
    ) -> ObjectImplementationSerializer:
        if not isinstance(impl, impl_mod.Data):
            return None
        value_encoder = registry.get_encoder(type, serializable=True)
        value_type_serializer = None
        if impl.value_type is not None:
            value_type_serializer = registry.get_type_serializer(impl.value_type)

        type_serializer = registry.get_type_serializer(type)
        return cls(value_encoder, type_serializer, value_type_serializer)

    @classmethod
    @st.hookimpl
    def get_impl_serializer_from_data(
        cls, data: Any, registry: "Registry"
    ) -> ObjectImplementationSerializer:
        if not isinstance(data, dict) or data.get("name") != "data":
            return None
        type_ser = registry.get_type_serializer_from_data(data["type"])
        typ = type_ser.deserialize(data["type"])
        value_encoder = registry.get_encoder(typ)

        value_type_ser = None
        if data["value_type"] is not None:
            value_type_ser = registry.get_type_serializer_from_data(data["value_type"])

        return cls(value_encoder, type_ser, value_type_ser)


@dc.dataclass(frozen=True)
class DataArraySerializer(DataSerializer):
    """
    Data serializer for array types
    """

    encoder: encoders.Encoder
    type_serializer: ts.TypeSerializer
    element_serializer: ObjectImplementationSerializer
    value_type_serializer: Optional[ts.TypeSerializer] = None

    def serialize(self, implementation: impl.ObjectImplementation) -> Any:
        out = {"type": self.type_serializer.serialize(self.encoder.type)}
        if isinstance(implementation.value, Object):
            out["value"] = {"object": implementation.value._impl.id}
        else:
            data = []
            if implementation.value is None:
                data = None
            else:
                encoded_value = self.encoder.encode(implementation.value)
                for item in encoded_value:
                    if isinstance(item, Object):
                        data.append({"object": item._impl.id})
                    else:
                        new_impl = impl.Data(item)
                        serialized = self.element_serializer.serialize(new_impl)[
                            "value"
                        ]
                        data.append(serialized)

            out["value"] = {"data": data}

        if (
            implementation.value_type is not None
            and self.value_type_serializer is not None
        ):
            out["value_type"] = self.value_type_serializer.serialize(
                implementation.value_type
            )

        return DataSchema().dump(out)

    def deserialize(
        self, data: Any, session: session.Session, objects: Mapping[Any, Object]
    ) -> impl.ObjectImplementation:

        deserialized = DataSchema().load(data)
        val_json = deserialized["value"]
        if "object" in val_json:
            value = objects[val_json["object"]]
        elif "data" in val_json:
            value_list = val_json["data"]
            value = []
            if value_list is None:
                value = None
            else:
                for item in value_list:
                    if "object" in item:
                        item_value = objects[item["object"]]
                    elif "data" in item:
                        sub_impl = self.element_serializer.deserialize(
                            {"value": item["data"]}, session, objects
                        )
                        item_value = sub_impl.value
                    else:
                        raise ValueError(f"Unable to interpret {val_json}!")

                    value.append(item_value)
        else:
            raise ValueError(f"Unable to interpret {val_json}!")

        value_type = None
        if (
            deserialized["value_type"] is not None
            and self.value_type_serializer is not None
        ):
            value_type = self.value_type_serializer.deserialize(
                deserialized["value_type"]
            )

        return impl.Data(value, value_type)

    @classmethod
    @st.hookimpl
    def get_impl_serializer(
        cls, impl: impl.ObjectImplementation, type: types.Type, registry: "Registry"
    ) -> ObjectImplementationSerializer:
        if not isinstance(impl, impl_mod.Data):
            return None
        if not isinstance(type, types.ArrayType):
            return None
        encoder = registry.get_encoder(type, serializable=True)
        element_serializer = registry.get_impl_serializer(
            impl=impl_mod.Data(None), type=type.element_type
        )

        value_type_serializer = None
        if impl.value_type is not None:
            value_type_serializer = registry.get_type_serializer(impl.value_type)

        type_serializer = registry.get_type_serializer(type)
        return cls(encoder, type_serializer, element_serializer, value_type_serializer)

    @classmethod
    @st.hookimpl
    def get_impl_serializer_from_data(
        cls, data: Any, registry: "Registry"
    ) -> ObjectImplementationSerializer:
        if not isinstance(data, dict) or data.get("name") != "data":
            return None

        type_ser = registry.get_type_serializer_from_data(data["type"])
        typ = type_ser.deserialize(data["type"])
        if not isinstance(typ, types.ArrayType):
            return None

        encoder = registry.get_encoder(typ)
        element_type_ser = registry.get_type_serializer(typ.element_type)

        element_data = {
            "name": "data",
            "value": None,
            "type": element_type_ser.serialize(typ.element_type),
            "value_type": None,
        }
        element_serializer = registry.get_impl_serializer_from_data(element_data)

        value_type_ser = None
        if data["value_type"] is not None:
            value_type_ser = registry.get_type_serializer_from_data(data["value_type"])

        return cls(encoder, type_ser, element_serializer, value_type_ser)


@dc.dataclass(frozen=True)
class DataMapSerializer(DataSerializer):
    """
    Data serializer for array types
    """

    encoder: encoders.Encoder
    type_serializer: ts.TypeSerializer
    key_serializer: ObjectImplementationSerializer
    value_serializer: ObjectImplementationSerializer
    value_type_serializer: Optional[ts.TypeSerializer] = None

    def serialize(self, implementation: impl.ObjectImplementation) -> Any:
        out = {"type": self.type_serializer.serialize(self.encoder.type)}
        if isinstance(implementation.value, Object):
            out["value"] = {"object": implementation.value._impl.id}
        else:
            data = []
            if implementation.value is None:
                data = None
            else:
                encoded_value = self.encoder.encode(implementation.value)
                for key, val in encoded_value.items():
                    if isinstance(key, Object):
                        key_dict = {"object": key._impl.id}
                    else:
                        new_impl = impl.Data(key)
                        key_dict = self.key_serializer.serialize(new_impl)["value"]

                    if isinstance(value, Object):
                        value_dict = {"object": value._impl.id}
                    else:
                        new_impl = impl.Data(value)
                        value_dict = self.value_serializer.serialize(new_impl)["value"]

                    data.append((key_dict, value_dict))

            out["value"] = {"data": data}

        if (
            implementation.value_type is not None
            and self.value_type_serializer is not None
        ):
            out["value_type"] = self.value_type_serializer.serialize(
                implementation.value_type
            )

        return DataSchema().dump(out)

    def deserialize(
        self, data: Any, session: session.Session, objects: Mapping[Any, Object]
    ) -> impl.ObjectImplementation:

        deserialized = DataSchema().load(data)
        val_json = deserialized["value"]
        if "object" in val_json:
            value = objects[val_json["object"]]
        elif "data" in val_json:
            value_list = val_json["data"]
            value = {}
            if value_list is None:
                value = None
            else:
                for key_dict, value_dict in value_list:
                    if "object" in key_dict:
                        key = objects[key_dict["object"]]
                    elif "data" in key_dict:
                        sub_impl = self.element_serializer.deserialize(
                            {"value": key_dict["data"]}, session, objects
                        )
                        key = sub_impl.value
                    else:
                        raise ValueError(f"Unable to interpret {key_dict}!")

                    if "object" in value_dict:
                        val = objects[value_dict["object"]]
                    elif "data" in value_dict:
                        sub_impl = self.element_serializer.deserialize(
                            {"value": value_dict["data"]}, session, objects
                        )
                        val = sub_impl.value
                    else:
                        raise ValueError(f"Unable to interpret {value_dict}!")

                    value[key] = val

        else:
            raise ValueError(f"Unable to interpret {val_json}!")

        value_type = None
        if (
            deserialized["value_type"] is not None
            and self.value_type_serializer is not None
        ):
            value_type = self.value_type_serializer.deserialize(
                deserialized["value_type"]
            )

        return impl.Data(value, value_type)

    @classmethod
    @st.hookimpl
    def get_impl_serializer(
        cls, impl: impl.ObjectImplementation, type: types.Type, registry: "Registry"
    ) -> ObjectImplementationSerializer:
        if not isinstance(impl, impl_mod.Data):
            return None
        if not isinstance(type, types.ArrayType):
            return None
        encoder = registry.get_encoder(type, serializable=True)
        element_serializer = registry.get_impl_serializer(
            impl=impl_mod.Data(None), type=type.element_type
        )

        value_type_serializer = None
        if impl.value_type is not None:
            value_type_serializer = registry.get_type_serializer(impl.value_type)

        type_serializer = registry.get_type_serializer(type)
        return cls(encoder, type_serializer, element_serializer, value_type_serializer)

    @classmethod
    @st.hookimpl
    def get_impl_serializer_from_data(
        cls, data: Any, registry: "Registry"
    ) -> ObjectImplementationSerializer:
        if not isinstance(data, dict) or data.get("name") != "data":
            return None

        type_ser = registry.get_type_serializer_from_data(data["type"])
        typ = type_ser.deserialize(data["type"])
        if not isinstance(typ, types.ArrayType):
            return None

        encoder = registry.get_encoder(typ)
        element_type_ser = registry.get_type_serializer(typ.element_type)

        element_data = {
            "name": "data",
            "value": None,
            "type": element_type_ser.serialize(typ.element_type),
            "value_type": None,
        }
        element_serializer = registry.get_impl_serializer_from_data(element_data)

        value_type_ser = None
        if data["value_type"] is not None:
            value_type_ser = registry.get_type_serializer_from_data(data["value_type"])

        return cls(encoder, type_ser, element_serializer, value_type_ser)


@dc.dataclass(frozen=True)
class DataStructSerializer(DataSerializer):
    """
    Data serializer for struct types
    """

    encoder: encoders.Encoder
    type_serializer: ts.TypeSerializer
    field_serializers: Dict[str, ObjectImplementationSerializer]
    value_type_serializer: Optional[ts.TypeSerializer] = None

    def serialize(self, implementation: impl.ObjectImplementation) -> Any:
        out = {"type": self.type_serializer.serialize(self.encoder.type)}
        if isinstance(implementation.value, Object):
            out["value"] = {"object": implementation.value._impl.id}
        else:
            data = {}
            if implementation.value is None:
                data = None
            else:
                encoded_value = self.encoder.encode(implementation.value)

                for field, value in encoded_value.items():
                    if isinstance(value, Object):
                        data[field] = {"object": value._impl.id}
                    else:
                        serializer = self.field_serializers[field]
                        new_impl = impl.Data(value)
                        serialized = serializer.serialize(new_impl)["value"]
                        data[field] = serialized

            out["value"] = {"data": data}

        if (
            implementation.value_type is not None
            and self.value_type_serializer is not None
        ):
            out["value_type"] = self.value_type_serializer.serialize(
                implementation.value_type
            )

        return DataSchema().dump(out)

    def deserialize(
        self, data: Any, session: session.Session, objects: Mapping[Any, Object]
    ) -> impl.ObjectImplementation:

        deserialized = DataSchema().load(data)
        val_json = deserialized["value"]
        if "object" in val_json:
            value = objects[val_json["object"]]
        elif "data" in val_json:
            value_dict = val_json["data"]
            value = {}
            if value_dict is None:
                value = None
            else:
                for field, item in value_dict.items():
                    if "object" in item:
                        item_value = objects[item["object"]]
                    elif "data" in item:
                        serializer = self.field_serializers[field]
                        sub_data = {"value": {"data": item["data"]}, "name": "data"}
                        sub_impl = serializer.deserialize(sub_data, session, objects)
                        item_value = sub_impl.value
                    else:
                        raise ValueError(f"Unable to interpret {val_json}!")

                    value[field] = item_value

        else:
            raise ValueError(f"Unable to interpret {val_json}!")

        value_type = None
        if (
            deserialized["value_type"] is not None
            and self.value_type_serializer is not None
        ):
            value_type = self.value_type_serializer.deserialize(
                deserialized["value_type"]
            )

        return impl.Data(value, value_type)

    @classmethod
    @st.hookimpl
    def get_impl_serializer(
        cls, impl: impl.ObjectImplementation, type: types.Type, registry: "Registry"
    ) -> ObjectImplementationSerializer:
        if not isinstance(impl, impl_mod.Data):
            return None
        if not isinstance(type, types.StructType):
            return None
        encoder = registry.get_encoder(type, serializable=True)
        field_serializers = {}
        for field in type.fields:
            field_serializers[field.name] = registry.get_impl_serializer(
                impl=impl_mod.Data(None), type=field.type
            )

        value_type_serializer = None
        if impl.value_type is not None:
            value_type_serializer = registry.get_type_serializer(impl.value_type)

        type_serializer = registry.get_type_serializer(type)
        return cls(encoder, type_serializer, field_serializers, value_type_serializer)

    @classmethod
    @st.hookimpl
    def get_impl_serializer_from_data(
        cls, data: Any, registry: "Registry"
    ) -> ObjectImplementationSerializer:
        if not isinstance(data, dict) or data.get("name") != "data":
            return None

        type_ser = registry.get_type_serializer_from_data(data["type"])
        typ = type_ser.deserialize(data["type"])
        if not isinstance(typ, types.StructType):
            return None

        encoder = registry.get_encoder(typ)

        field_serializers = {}
        for field in typ.fields:
            field_type_ser = registry.get_type_serializer(field.type)
            field_data = {
                "name": "data",
                "value": None,
                "type": field_type_ser.serialize(field.type),
                "value_type": None,
            }
            field_serializers[field.name] = registry.get_impl_serializer_from_data(
                field_data
            )

        value_type_ser = None
        if data["value_type"] is not None:
            value_type_ser = registry.get_type_serializer_from_data(data["value_type"])

        return cls(encoder, type_ser, field_serializers, value_type_ser)


class FunctionCallSchema(ma.Schema):
    """
    Output schema for a function call
    """

    name = ma.fields.Str(required=True, default="function_call")
    func = ma.fields.Dict()
    func_type = ma.fields.Dict()
    arguments = ma.fields.Dict(
        keys=ma.fields.Str(), values=ma.fields.Field(), missing=dict
    )


@dc.dataclass(frozen=True)
class FunctionCallSerializer(ObjectImplementationSerializer):
    """
    Serializer for a function call
    """

    function_encoder: encoders.Encoder
    function_type_serializer: ts.TypeSerializer

    def serialize(self, implementation: impl.ObjectImplementation) -> Any:
        encoded_func = self.function_encoder.encode(implementation.func)
        func_type = self.function_type_serializer.serialize(implementation.func.type)
        arguments = {key: obj._impl.id for key, obj in implementation.arguments.items()}

        out = {"func": encoded_func, "func_type": func_type, "arguments": arguments}
        return FunctionCallSchema().dump(out)

    def deserialize(
        self, data: Any, session: session.Session, objects: Mapping[Any, Object]
    ) -> impl.ObjectImplementation:

        loaded_data = FunctionCallSchema().load(data)
        loaded_func = self.function_encoder.decode(loaded_data["func"])
        loaded_kwargs = {
            key: objects[obj_id] for key, obj_id in loaded_data["arguments"].items()
        }
        return impl.FunctionCall(loaded_func, loaded_kwargs)

    @classmethod
    @st.hookimpl
    def get_impl_serializer(
        cls, impl: impl.ObjectImplementation, type: types.Type, registry: "Registry"
    ) -> ObjectImplementationSerializer:

        if not isinstance(impl, impl_mod.FunctionCall):
            return None
        func_encoder = registry.get_encoder(impl.func.type)
        func_type_ser = registry.get_type_serializer(impl.func.type)
        return cls(func_encoder, func_type_ser)

    @classmethod
    @st.hookimpl
    def get_impl_serializer_from_data(
        cls, data: Any, registry: "Registry"
    ) -> ObjectImplementationSerializer:
        if not isinstance(data, dict) or data.get("name") != "function_call":
            return None

        func_type_ser = registry.get_type_serializer_from_data(data["func_type"])
        func_type = func_type_ser.deserialize(data["func_type"])
        func_encoder = registry.get_encoder(func_type)

        return cls(func_encoder, func_type_ser)


IMPL_SERIALIZER_CLASSES = [
    ReferenceSerializer,
    FutureSerializer,
    ExpectedValueSerializer,
    DataValueSerializer,
    DataArraySerializer,
    DataStructSerializer,
    FunctionCallSerializer,
]


def register(registry: Optional["Registry"] = None) -> None:
    """
    Register casters in the default registry
    """
    if registry is None:
        registry = st.registry

    for plugin in IMPL_SERIALIZER_CLASSES:
        registry.register(plugin)



import abc
import dataclasses as dc
from typing import Any, Optional, Mapping

import statey as st
from statey.syms import Object, type_serializers, impl_serializers, session


class ObjectSerializer(abc.ABC):
    """
    Serializes an object
    """

    @abc.abstractmethod
    def serialize(self, obj: Object) -> Any:
        """
        Serialize an object into a JSON-serializable object
        """
        raise NotImplementedError

    @abc.abstractmethod
    def deserialize(
        self, data: Any, session: session.Session, objects: Mapping[Any, Object]
    ) -> Object:
        """
        Deserialize the given data into an object
        """
        raise NotImplementedError


@dc.dataclass(frozen=True)
class DefaultObjectSerializer(abc.ABC):
    """
    Default object serializer implementation
    """

    type_serializer: type_serializers.TypeSerializer
    impl_serializer: impl_serializers.ObjectImplementationSerializer

    def serialize(self, obj: Object) -> Any:
        serialized_type = self.type_serializer.serialize(obj._type)
        serialized_impl = self.impl_serializer.serialize(obj._impl)
        return {"type": serialized_type, "impl": serialized_impl}

    def deserialize(
        self, data: Any, session: session.Session, objects: Mapping[Any, Object]
    ) -> Object:
        typ = self.type_serializer.deserialize(data["type"])
        impl = self.impl_serializer.deserialize(data["impl"], session, objects)
        return Object(impl, typ, session.ns.registry)

    @classmethod
    @st.hookimpl
    def get_object_serializer(
        cls, obj: "Object", registry: "Registry"
    ) -> ObjectSerializer:
        type_serializer = registry.get_type_serializer(obj._type)
        impl_serializer = registry.get_impl_serializer(obj._impl, obj._type)
        return cls(type_serializer, impl_serializer)

    @classmethod
    @st.hookimpl
    def get_object_serializer_from_data(
        cls, data: Any, registry: "Registry"
    ) -> ObjectSerializer:
        impl_serializer = registry.get_impl_serializer_from_data(data["impl"])
        type_serializer = registry.get_type_serializer_from_data(data["type"])
        return cls(type_serializer, impl_serializer)


OBJECT_SERIALIZER_CLASSES = [DefaultObjectSerializer]


def register(registry: Optional["Registry"] = None) -> None:
    """
    Register all object serializers
    """
    if registry is None:
        registry = st.registry

    for cls in OBJECT_SERIALIZER_CLASSES:
        registry.register(cls)



import abc
import dataclasses as dc
from functools import wraps
from itertools import zip_longest
from typing import Iterable, Any, Optional, Sequence, Dict

import networkx as nx

from statey import exc
from statey.syms import base, func, utils, types, stack
from statey.syms.object_ import Object


class ObjectImplementation(base.AttributeAccess):
    """
    Some implementation for an object
    """

    @abc.abstractmethod
    def depends_on(self, obj: Object, session: "Session") -> Iterable[Object]:
        """
        Retrieve any objects that this one depends on
        """
        raise NotImplementedError

    @abc.abstractmethod
    def apply(self, obj: Object, dag: nx.DiGraph, session: "Session") -> Any:
        """
        Given the dag containing results of any upstreams as well
        as the current session, fully resolve the value of this object
        """
        raise NotImplementedError

    @abc.abstractmethod
    def map(self, obj: Object, function: func.Function) -> Object:
        """
        Apply the given function and return a new object.
        """
        raise NotImplementedError

    def apply_alone(self, obj: Object) -> Any:
        """
        Object implementations may not depend on other objects and thus can
        be trivially resolved this way. Default behavior is to return NotImplemented
        """
        raise NotImplementedError

    # def call(self, obj: Object, args: Sequence[Any], kwargs: Dict[str, Any]) -> Object:
    #     """
    #     Some object implementations can apply arguments
    #     """
    #     raise NotImplementedError

    @property
    def id(self) -> Any:
        """
        Get some unique identifier for this object implementation
        """
        return id(self)

    def object_repr(self, obj: Object) -> str:
        """
        Render a string representation of an object with this implementation
        """
        return f"{type(obj).__name__}({repr(self)})"
        # return f"{type(obj).__name__}[{obj._type}]({repr(self)})"

    def type(self) -> types.Type:
        """
        Some object implementations are bound to a specific type, so this method
        allows the ability to infer the type from the implementatino. Returning
        NotImplemented incidates the type is not necessarily known
        """
        raise NotImplementedError

    def registry(self) -> "Registry":
        """
        Some object implementations are bound to a specific registry, so this method
        allows the ability to infer the registry from the implementation. Default
        behavior is returning NotImplemented
        """
        raise NotImplementedError

    def frame(self) -> stack.FrameSnapshot:
        """
        Allow an object frame to be inferred from an implementation
        """
        raise NotImplementedError


class FunctionalAttributeAccessMixin:
    """
    Get attributes by simplying applying semantic get_attr function within
    """

    def get_attr(self, obj: Object, attr: str) -> Any:
        from statey.syms import api

        semantics = obj._registry.get_semantics(obj._type)
        attr_semantics = semantics.attr_semantics(attr)

        if attr_semantics is None:
            raise exc.SymbolAttributeError(obj, attr)

        getter_func = lambda x: semantics.get_attr(x, attr)
        func_type = utils.single_arg_function_type(
            semantics.type, attr_semantics.type, "x"
        )
        function_obj = api.function(getter_func, func_type)
        return self.map(obj, function_obj)


class FunctionalMappingMixin:
    """
    Defines default map() behavior that works for almost all cases
    """

    def map(self, obj: Object, function: func.Function) -> Object:
        return utils.wrap_function_call(function, (obj,), registry=obj._registry)


class FunctionalBehaviorMixin(FunctionalMappingMixin, FunctionalAttributeAccessMixin):
    """
    Simple way to mix in all default functional behavior for object impls
    """


@dc.dataclass(frozen=True)
class Reference(FunctionalMappingMixin, ObjectImplementation):
    """
    References some key in a session, possibly with some relative path
    """

    path: str
    ns: "Namespace" = dc.field(repr=False, hash=False, compare=False)

    def get_attr(self, obj: Object, attr: str) -> Any:
        semantics = obj._registry.get_semantics(obj._type)
        path = self.ns.path_parser.join([self.path, attr])
        if semantics is None:
            raise exc.SymbolKeyError(path, ns)

        new_ref = Reference(path, self.ns)
        return Object(new_ref, frame=obj._frame)

    def depends_on(self, obj: Object, session: "Session") -> Iterable[Object]:
        semantics = obj._registry.get_semantics(obj._type)
        data = session.get_encoded_data(self.path)
        expanded = semantics.expand(data)

        syms = []

        semantics.map_objects(syms.append, expanded)
        return syms

    def apply(self, obj: Object, dag: nx.DiGraph, session: "Session") -> Any:
        return session.get_encoded_data(self.path)

    @property
    def id(self) -> Any:
        return f"{type(self).__name__}:{self.path}"

    def type(self) -> types.Type:
        return self.ns.resolve(self.path)

    def registry(self) -> "Registry":
        return self.ns.registry

    def object_repr(self, obj: "Object") -> str:
        return f"{type(self).__name__}({self.path})"
        # return f"{type(self).__name__}[{obj._type}]({self.path})"


class StandaloneObjectImplementation(ObjectImplementation):
    """
    Object implementation for impementations that never have dependencies
    """

    def depends_on(self, obj: Object, session: "Session") -> Iterable[Object]:
        return ()

    def apply(self, obj: Object, dag: nx.DiGraph, session: "Session") -> Any:
        return self.apply_alone(obj)

    @abc.abstractmethod
    def apply_alone(self, obj: Object) -> Any:
        """
        This must be implemented in StandaloneObjectImplementation subclasses
        """
        raise NotImplementedError


@dc.dataclass(frozen=True)
class Data(FunctionalMappingMixin, StandaloneObjectImplementation):
    """
    Object implementation for concrete data
    """

    value: Any
    value_type: Optional[types.Type] = None
    encoded_value: Any = utils.MISSING
    encoded_value_type: Optional[types.Type] = None

    def get_encoded_value(self, type: types.Type, registry: "Registry") -> Any:
        """
        Get the encoded version of this value, calculating it if
        it hasn't been
        """
        # We only want to encode the value once, but we also need to be
        # careful to encode the value with the encoder from the correct type,
        # including if `obj` is casted from one type to another (for example adding
        # a validator)
        if utils.is_missing(self.encoded_value) or type != self.encoded_value_type:
            encoder = registry.get_encoder(type)
            semantics = registry.get_semantics(type)
            encoded_value = encoder.encode(self.value)
            expanded_value = semantics.expand(encoded_value)

            self.__dict__["encoded_value"] = expanded_value
            self.__dict__["encoded_value_type"] = type

        return self.encoded_value

    def depends_on(self, obj: Object, session: "Session") -> Iterable[Object]:
        expanded = self.get_encoded_value(obj._type, obj._registry)

        syms = []
        semantics = obj._registry.get_semantics(obj._type)
        semantics.map_objects(syms.append, expanded)
        return syms

    def apply_alone(self, obj: Object) -> Any:
        return self.get_encoded_value(obj._type, obj._registry)

    def type(self) -> types.Type:
        if self.value_type is None:
            raise NotImplementedError
        return self.value_type

    def object_repr(self, obj: "Object") -> str:
        return f"{type(self).__name__}({repr(self.value)})"
        # return f"{type(self).__name__}[{obj._type}]({repr(self.value)})"

    def get_attr(self, obj: Object, attr: str) -> Any:
        encoded_value = self.get_encoded_value(obj._type, obj._registry)
        semantics = obj._registry.get_semantics(obj._type)
        if encoded_value is None:
            new_data = None
        else:
            new_data = semantics.get_attr(encoded_value, attr)
        attr_semantics = semantics.attr_semantics(attr)
        if attr_semantics is None:
            raise exc.SymbolAttributeError(obj, attr)

        new_impl = new_data
        attr_type = attr_semantics.type
        if not isinstance(new_data, Object):
            new_impl = Data(new_data, attr_type, new_data, attr_type)

        # Data is already encoded, can pass the encoded values explicitly
        return Object(new_impl, attr_type, registry=obj._registry, frame=obj._frame)


@dc.dataclass(frozen=True)
class FunctionCall(FunctionalBehaviorMixin, ObjectImplementation):
    """
    Object implementation for a function and associated arguments
    """

    func: func.Function
    arguments: Dict[str, Object] = None

    def __post_init__(self) -> None:
        if self.arguments is None:
            self.__dict__["arguments"] = {}

    def depends_on(self, obj: Object, session: "Session") -> Iterable[Object]:
        yield from self.arguments.values()

    def apply(self, obj: Object, dag: nx.DiGraph, session: "Session") -> Any:
        unknowns = []

        kwargs = {}
        for key, sym in self.arguments.items():
            if not isinstance(sym, Object):
                kwargs[key] = sym
                continue
            arg = dag.nodes[sym._impl.id]["result"]
            arg_encoder = session.ns.registry.get_encoder(sym._type)
            decoded_arg = arg_encoder.decode(arg)
            semantics = obj._registry.get_semantics(sym._type)
            semantics.map_objects(unknowns.append, arg)
            kwargs[key] = decoded_arg

        if unknowns:
            raise exc.UnknownError(unknowns)

        result = self.func.apply(kwargs)
        # Not using return type to account for casting
        encoder = session.ns.registry.get_encoder(obj._type)
        return encoder.encode(result)

    def type(self) -> types.Type:
        return self.func.type.return_type

    def object_repr(self, obj: "Object") -> str:
        kwarg_reprs = ", ".join(
            "=".join([key, repr(val)]) for key, val in self.arguments.items()
        )
        return f"{type(self.func).__name__}Call({self.func.name}({kwarg_reprs}))"
        # return f"{type(self.func).__name__}Call[{obj._type}]({self.func.name}({kwarg_reprs}))"


@dc.dataclass(frozen=True)
class FutureResult(utils.Cloneable):
    """
    Helper so that we can clone futures while still retaining their behavior.
    """

    result: Any = dc.field(default=utils.MISSING)

    def get(self) -> Any:
        """
        Get the result of the future, raising exc.FutureResultNotSet
        if it hasn't been set yet
        """
        if self.result is utils.MISSING:
            raise exc.FutureResultNotSet(self)
        return self.result

    def set(self, result: Any, overwrite: bool = False) -> None:
        """
        Set the result, raising exc.FutureResultAlreadySet if it has
        already been set
        """
        if self.result is not utils.MISSING and not overwrite:
            raise exc.FutureResultAlreadySet(self)
        self.__dict__["result"] = result


@dc.dataclass(frozen=True)
class Future(FunctionalBehaviorMixin, ObjectImplementation):
    """
    A future is a value that may or may not yet be set
    """

    refs: Sequence[Object] = ()
    result: FutureResult = dc.field(default_factory=FutureResult)
    return_type: Optional[types.Type] = None

    def get_result(self) -> Any:
        """
        Get the result of the future, raising exc.FutureResultNotSet
        if it hasn't been set yet
        """
        return self.result.get()

    def set_result(self, result: Any, overwrite: bool = False) -> None:
        """
        Set the result, raising exc.FutureResultAlreadySet if it has
        already been set
        """
        self.result.set(result, overwrite)

    def depends_on(self, obj: Object, session: "Session") -> Iterable[Object]:
        return self.refs

    def apply(self, obj: Object, dag: nx.DiGraph, session: "Session") -> Any:
        try:
            res = self.apply_alone(obj)
        except NotImplementedError as err:
            raise exc.UnknownError from err
        return res

    def apply_alone(self, obj: Object) -> Any:
        try:
            return self.get_result()
        except exc.FutureResultNotSet as err:
            raise NotImplementedError from err

    def type(self) -> types.Type:
        if self.return_type is None:
            raise NotImplementedError
        return self.return_type

    def object_repr(self, obj: Object) -> str:
        return f"{type(self).__name__}({self.result.result})"
        # return f"{type(self).__name__}[{obj._type}]({self.result.result})"


@dc.dataclass(frozen=True)
class Unknown(ObjectImplementation):
    """
    Some value that is not known
    """

    def __class_getitem__(self, item: Any) -> Object:
        """
        Get an Unknown object with the given type
        """
        import statey as st

        typ = st.registry.get_type(item)
        return st.Object(Unknown(return_type=typ))

    obj: Optional[Object] = None
    refs: Sequence[Object] = ()
    return_type: Optional[types.Type] = None

    def __post_init__(self) -> None:
        return_type = self.return_type
        while self.obj is not None and isinstance(self.obj._impl, Unknown):
            return_type = self.obj._impl.return_type
            self.__dict__["obj"] = self.obj._impl.obj

        if self.return_type is None:
            if return_type is not None:
                self.__dict__["return_type"] = return_type
            elif self.obj is not None:
                self.__dict__["return_type"] = self.obj._type

    def depends_on(self, obj: Object, session: "Session") -> Iterable[Object]:
        return self.refs

    def apply(self, obj: Object, dag: nx.DiGraph, session: "Session") -> Any:
        raise exc.UnknownError(self.refs)

    def get_attr(self, obj: Object, attr: str) -> Any:
        def handle(result):

            if isinstance(result, Object):
                new_impl = Unknown(result, self.refs)
                return Object(new_impl, result._type, result._registry)

            if callable(result):

                @wraps(result)
                def wrapper(*args, **kwargs):
                    return handle(result(*args, **kwargs))

                return wrapper

            raise TypeError(f"Unhandled attribute result type {result}! Failing")

        if self.obj is None:
            semantics = obj._registry.get_semantics(obj._type)
            attr_semantics = semantics.attr_semantics(attr)
            if attr_semantics is None:
                raise exc.SymbolAttributeError(obj, attr)

            new_impl = Unknown(refs=self.refs, return_type=attr_semantics.type)
            return Object(new_impl, attr_semantics.type, obj._registry)

        return handle(self.obj[attr])

    def map(self, obj: Object, function: func.Function) -> Object:

        if self.obj is None:
            new_impl = Unknown(refs=self.refs, return_type=function.type.return_type)
            return Object(new_impl, function.type.return_type, obj._registry)

        mapped_object = self.obj._inst.map(function)
        new_impl = Unknown(mapped_object, self.refs)
        return Object(new_impl, mapped_object._type, mapped_object._registry)

    def type(self) -> types.Type:
        if self.return_type is None:
            return NotImplementedError
        return self.return_type

    def registry(self) -> "Registry":
        if self.obj is None:
            raise NotImplementedError
        return self.obj._registry

    def object_repr(self, obj: "Object") -> str:
        post = "" if self.obj is None else f"({self.obj})"
        return f"{type(self).__name__}{post}"
        # return f"{type(self).__name__}[{obj._type}]{post}"


@dc.dataclass(frozen=True)
class StructField:
    """
    Single field in a StructSymbol
    """

    name: str
    value: Object


@dc.dataclass(frozen=True)
class Struct(FunctionalMappingMixin, ObjectImplementation):
    """
    Combines multiple objects into a struct
    """

    fields: Sequence[StructField]

    def get_attr(self, obj: Object, attr: str) -> Any:
        field_map = {field.name: field for field in self.fields}
        return field_map[attr].value

    def depends_on(self, obj: Object, session: "Session") -> Iterable[Object]:
        for field in self.fields:
            yield from field.value._impl.depends_on(obj[field.name], session)

    def apply(self, obj: Object, dag: nx.DiGraph, session: "Session") -> Any:
        out = {}
        for field in self.fields:
            out[field.name] = field.value._impl.apply(obj[field.name], dag, session)
        return out

    def type(self) -> types.Type:
        fields = []
        for field in self.fields:
            typ = field.value._type
            fields.append(types.Field(field.name, typ))
        return types.StructType(fields, False)

    def object_repr(self, obj: "Object") -> str:
        field_reprs = ", ".join(
            "=".join([field.name, repr(field.value)]) for field in self.fields
        )
        return f"{type(self).__name__}({field_reprs})"
        # return f"{type(self).__name__}[{obj._type}]({field_reprs})"


@dc.dataclass(frozen=True)
class ExpectedValue(ObjectImplementation):
    """
    Defines some parent object and an expectation of what the value will be after resolved
    """

    obj: Object
    expected: Object

    def __post_init__(self) -> None:
        assert isinstance(self.expected, Object)

    # def get_attr(self, obj: Object, attr: str) -> Any:
    #     new_impl = ExpectedValue(self.obj[attr], self.expected[attr])
    #     return Object(new_impl)

    def get_attr(self, obj: Object, attr: str) -> Any:
        def handle(result):

            if isinstance(result, Object):
                new_impl = ExpectedValue(result, self.expected[attr])
                return Object(
                    new_impl, result._type, result._registry, frame=obj._frame
                )

            if callable(result):

                @wraps(result)
                def wrapper(*args, **kwargs):
                    res = result(*args, **kwargs)
                    expected_res = self.expected[attr](*args, **kwargs)
                    new_impl = ExpectedValue(res, expected_res)
                    return Object(
                        new_impl,
                        res._type,
                        res._registry,
                        frame=stack.frame_snapshot(1),
                    )

                return wrapper

            raise TypeError(f"Unhandled attribute result type {result}! Failing")

        return handle(self.obj[attr])

    def depends_on(self, obj: Object, session: "Session") -> Iterable[Object]:
        yield from self.obj._impl.depends_on(self.obj, session)

    def apply(self, obj: Object, dag: nx.DiGraph, session: "Session") -> Any:
        try:
            return self.obj._impl.apply(self.obj, dag, session)
        except exc.UnknownError as err:
            refs = list(self.depends_on(obj, session))
            expected_value = self.expected
            try:
                expected_value = session.resolve(self.expected, allow_unknowns=True)
            except Exception:
                pass
            raise exc.UnknownError(refs, expected=expected_value) from err

    def map(self, obj: Object, function: func.Function) -> Object:
        mapped_obj = self.obj._inst.map(function)
        mapped_expected = self.expected._inst.map(function)
        new_impl = ExpectedValue(mapped_obj, mapped_expected)
        return Object(new_impl, function.type.return_type, obj._registry)

    def type(self) -> types.Type:
        return self.obj._type

    def registry(self) -> "Registry":
        return self.obj._registry

    def object_repr(self, obj: "Object") -> str:
        return f"{type(self).__name__}({self.expected}, from={self.obj})"
        # return f"{type(self).__name__}[{obj._type}]({self.expected}, from={self.obj})"



import dataclasses as dc
import os
import sys
from importlib.machinery import ModuleSpec

import jsonschema
import statey as st
import pylumi
from typing import Sequence, Dict, Any, Optional

from statey.ext.pulumi import __file__ as root_file
from statey.ext.pulumi.constants import RESOLVER_STORE, PULUMI_NS


ROOT_DIR = os.path.abspath(os.path.dirname(root_file))


@dc.dataclass(frozen=True)
class PulumiResourceSchema:
    """
    Describes
    """

    description: str
    input_type: st.Type
    output_type: st.Type


@dc.dataclass(frozen=True)
class PulumiProviderSchema:
    """
    Object representing a pulumi provider schema response

    (not all information consumed)
    """

    name: str
    version: str
    description: str
    resources: Dict[str, PulumiResourceSchema]


class PulumiProviderSchemaParser:
    """
    Parses a pulumi provider get_schema() response into a PulumiProviderSchema object
    """

    def _fix_broken_refs(self, doc: Dict[str, Any]) -> Dict[str, Any]:
        """
        In Pulumi schemas some keys contain slashes; this messes w/ jsonschema.RefResolver
        """
        keys = ["types", "resources", "functions"]

        out = {}

        for key in keys:

            doc_types = doc[key]
            this_out = out.setdefault(key, {})

            for key, val in list(doc_types.items()):
                comps = key.split("/")
                current = this_out
                for comp in comps[:-1]:
                    current = current.setdefault(comp, {})

                current[comps[-1]] = val

        return out

    def _resolve_refs(
        self, schema: Dict[str, Any], doc: Dict[str, Any]
    ) -> Dict[str, Any]:
        """
        Resolve any references in the given pulumi provider schema
        """
        if "$ref" in schema:
            resolver = jsonschema.RefResolver("", doc, RESOLVER_STORE)
            try:
                resolved = resolver.resolve(schema["$ref"])[1]
            except jsonschema.exceptions.RefResolutionError:
                second = {"type": "string"}
                ct2 = 0
            else:
                second, ct2 = self._resolve_refs(resolved, doc)
            return second, ct2 + 1

        resolved = 0
        schema_type = schema.get("type")
        if schema_type == "object":
            schema = schema.copy()

            if "properties" in schema:
                props = {}

                for key, val in schema.get("properties", {}).items():
                    out, ct = self._resolve_refs(val, doc)
                    props[key] = out
                    resolved += ct
                schema = dict(schema, properties=props)

            if "additionalProperties" in schema:
                schema["additionalProperties"], ct = self._resolve_refs(
                    schema["additionalProperties"], doc
                )
                resolved += ct

            if resolved > 0:
                second, ct2 = self._resolve_refs(schema, doc)
                return second, resolved + ct2
            return schema, resolved

        if schema_type == "array":
            out, ct = self._resolve_refs(schema["items"], doc)
            resolved += ct
            out_schema = dict(schema, items=out)
            if resolved > 0:
                return self._resolve_refs(out_schema, doc)
            return out_schema, resolved

        if schema_type is not None:
            return schema, resolved

        one_of = schema.get("oneOf")
        if one_of is None:
            return schema, resolved

        outs = []
        for item in one_of:
            out, ct = self._resolve_refs(item, doc)
            if out not in outs:
                outs.append(out)
                resolved += ct

        out_schema = schema.copy()
        del out_schema["oneOf"]
        out_schema.update(outs[-1])
        return out_schema, resolved

    def parse(
        self, data: Dict[str, Any], registry: Optional[st.Registry] = None
    ) -> PulumiProviderSchema:
        """
        Parse a dictionary response into a PulumiProviderSchema
        """
        if registry is None:
            registry = st.registry

        resources = data["resources"].copy()

        fixed_data = self._fix_broken_refs(data)

        out = {}
        for key, schema in resources.items():
            input_type = {
                "type": "object",
                "properties": schema.get("inputProperties", {}),
                "required": schema.get("requiredInputs", []),
            }
            input_type, _ = self._resolve_refs(input_type, fixed_data)
            input_ser = registry.get_type_serializer_from_data(input_type)
            input_type_obj = input_ser.deserialize(input_type)

            output_type = {
                "type": "object",
                "properties": schema.get("properties", {}),
                "required": schema.get("required", []),
            }
            output_type, _ = self._resolve_refs(output_type, fixed_data)
            output_ser = registry.get_type_serializer_from_data(output_type)
            output_type_obj = output_ser.deserialize(output_type)

            out[key] = PulumiResourceSchema(
                description=schema.get("description"),
                input_type=input_type_obj,
                output_type=output_type_obj,
            )

        return PulumiProviderSchema(
            name=data.get("name"),
            version=data.get("version"),
            description=data.get("description"),
            resources=out,
        )


def object_to_pulumi_json(obj: st.Object, session: st.Session) -> str:
    """
    Encode an object as JSON to transmit to pulumi
    """

    def convert_object(obj):
        typ = obj._type
        unknown_value = pylumi.UnknownValue.NULL_
        if isinstance(typ, st.StringType):
            unknown_value = pylumi.UnknownValue.STRING
        elif isinstance(typ, st.NumberType):
            unknown_value = pylumi.UnknownValue.NUMBER
        elif isinstance(typ, st.BooleanType):
            unknown_value = pylumi.UnknownValue.BOOL
        elif isinstance(typ, st.ArrayType):
            unknown_value = pylumi.UnknownValue.ARRAY
        elif isinstance(typ, (st.StructType, st.MapType)):
            unknown_value = pylumi.UnknownValue.OBJECT

        return unknown_value

    semantics = obj._registry.get_semantics(obj._type)
    data = session.resolve(obj, allow_unknowns=True)
    expanded = semantics.expand(data)
    with_encoded_unknowns = semantics.map_objects(convert_object, expanded)

    return with_encoded_unknowns


def parse_provider_schema_response(
    data: Dict[str, Any], registry: Optional[st.Registry] = None
) -> PulumiProviderSchema:
    """
    Parse a JSON schema response into a PulumiProviderSchema object
    """
    return PulumiProviderSchemaParser().parse(data, registry)


class ProviderFinder:
    """
    Module meta finder to allow loading providers via the import system
    """

    prefix = "statey.ext.pulumi.providers"

    def find_spec(self, fullname, path, target=None) -> ModuleSpec:
        from statey.ext.pulumi import provider_api

        if not path:
            return None
        if os.path.abspath(path[0]) != ROOT_DIR:
            return None
        if not fullname.startswith(self.prefix):
            return None
        name = fullname[len(self.prefix) + 1 :]
        if not name:
            return ModuleSpec(fullname, ProvidersLoader(), is_package=False)

        head, *path = name.split(".", 1)
        if path:
            return None

        loader = ProviderLoader(name)
        return ModuleSpec(fullname, loader, is_package=False)


class ProviderLoader:
    """
    Module loader to load a provider
    """

    def __init__(self, name: str) -> None:
        self.name = name

    def load_module(self, spec):
        from statey.ext.pulumi import provider_api

        api = sys.modules[spec] = provider_api.provider_api(self.name)
        api.__file__ = provider_api.__file__
        api.__name__ = spec
        api.__loader__ = self
        api.__package__ = spec.rsplit(".", 1)[0]
        return api


class ProvidersLoader:
    """
    Module loader to load the `providers` module
    """

    def load_module(self, spec):
        from statey.ext.pulumi import provider_api

        api = sys.modules[spec] = provider_api.ProvidersAPI()
        api.__file__ = provider_api.__file__
        api.__name__ = spec
        api.__loader__ = self
        api.__package__ = spec.rsplit(".", 1)[0]
        return api


FINDER = ProviderFinder()


def register_meta_finder() -> None:
    """
    Register a ProviderFinder in the import system
    """
    if FINDER not in sys.meta_path:
        sys.meta_path.append(FINDER)


def unregister_meta_finder() -> None:
    """
    Unregster the ProviderFinder from the meta path, if it is registered.
    """
    if FINDER in sys.meta_path:
        sys.meta_path.remove(FINDER)


class AwsEnvironmentHandler:
    """
    Plugins to allow for easy environment-based setup of the pulumi AWS provider
    """

    @staticmethod
    @st.hookimpl(hookwrapper=True)
    def get_provider_config(name, params, registry):
        if name == PULUMI_NS + "/aws":
            default_region = os.getenv("AWS_REGION", os.getenv("AWS_DEFAULT_REGION"))
            if default_region is not None:
                params.setdefault("region", default_region)

        outcome = yield


PLUGINS = [AwsEnvironmentHandler]


def register(registry: Optional[st.Registry] = None) -> None:
    """
    Register all plugins for this module
    """
    if registry is None:
        registry = st.registry

    register_meta_finder()

    for plugin in PLUGINS:
        registry.register(plugin)



#!/usr/bin/env python

from setuptools import setup, find_packages


with open('requirements.txt') as f:
	# Load one dependency per line, ignoring comments
	requirements = [
		line.strip()
		for line in f.read().strip().split('\n')
		if not line.strip().startswith('#')
	]


with open('requirements-tests.txt') as f:
	# Load one dependency per line, ignoring comments
	requirements_tests = [
		line.strip()
		for line in f.read().strip().split('\n')
		if not line.strip().startswith('#')
	]

with open('requirements-dev.txt') as f:
	# Load one dependency per line, ignoring comments
	requirements_dev = [
		line.strip()
		for line in f.read().strip().split('\n')
		if not line.strip().startswith('#')
	]

with open('README.rst') as f:
    long_description = f.read().strip()

extra_reqs = {
	'fmt': ['fmt==0.3.1'],
	'pickle': [
		'dill==0.3.2',
		'cloudpickle==1.4.1'
	],
	'cli': [
		'click==7.1.2',
        # TODO: This doesn't work w/ PyPI, need to either submit a PR
        # to the main repo or figure something else out
		# 'asciidag @ git+https://github.com/cfeenstra67/asciidag.git',
        'asciidag==0.2.0'
	],
	'pulumi': [
		'pylumi==1.2.2',
		'jsonschema==3.2.0'
	],
}

extra_reqs['all'] = sum(extra_reqs.values(), [])
extra_reqs['core'] = extra_reqs['fmt'] + extra_reqs['cli']
extra_reqs['tests'] = requirements_tests
extra_reqs['dev'] = requirements_dev

setup(
	name='statey',
	version='0.0.8',
	description='Graph-based provisioning framework.',
	long_description=long_description,
	long_description_content_type='text/x-rst',
	classifiers=[
        'Development Status :: 1 - Planning',
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License"
	],
	license='MIT',
	author='Cam Feenstra',
	author_email='cameron.l.feenstra@gmail.com',
	packages=find_packages(exclude=('tests', 'tests.*')),
	install_requires=requirements,
	extras_require=extra_reqs,
	entry_points={
		'console_scripts': [
			'statey=statey.cli.__main__:main'
		]
	},
	url='https://github.com/cfeenstra67/statey',
    package_data={
        '': [
            'requirements.txt',
            'requirements-dev.txt',
            'requirements-tests.txt',
            'README.rst',
        ],
    },
    include_package_data=True
)



"""
Modified from the Jinja code
"""
import contextlib
import dataclasses as dc
import platform
import sys
import textwrap as tw
from types import CodeType
from typing import Type as PyType, Any, Callable, Sequence, Optional


INTERNAL_CODE = set()


RESOLUTION_START = set()


@dc.dataclass(frozen=True)
class FrameSnapshot:
    """
    Marks the attributes of a frame at a particular point in time
    """

    lineno: int
    name: str
    filename: str


def frame_snapshot(depth: int = 0) -> FrameSnapshot:
    """
    Take a snapshot of the frame at the given depth (depth relative to the caller)
    """
    frame = sys._getframe(depth + 1)
    return FrameSnapshot(
        lineno=frame.f_lineno,
        filename=frame.f_code.co_filename,
        name=frame.f_code.co_name,
    )


def internalcode(func: Callable[[Any], Any]) -> Callable[[Any], Any]:
    """
    Mark the given function as internal (don't show in tracebacks when using)
    """
    INTERNAL_CODE.add(func.__code__)
    return func


def resolutionstart(func: Callable[[Any], Any]) -> Callable[[Any], Any]:
    """
    Mark the given function as the first frame before resolution of objects begins. Additional
    frames will be injected
    """
    RESOLUTION_START.add(func.__code__)
    return func


def default_stack_func(stack: Sequence[Any]) -> Sequence[Any]:
    """
    Default stack behavior, just hiding internal calls
    """
    out = []
    for tb in stack:
        if tb.tb_frame.f_code in INTERNAL_CODE:
            continue
        out.append(tb)
    return out


@internalcode
def rewrite_tb(
    exc_type: PyType[Exception],
    exc_value: Any,
    tb: Any,
    resolution_stack: Optional["ResolutionStack"] = None,
) -> None:
    """"""
    raise rewrite_traceback_stack(exc_type, exc_value, tb, resolution_stack)


@contextlib.contextmanager
@internalcode
def rewrite_ctx(resolution_stack: Optional["ResolutionStack"] = None):
    """"""
    try:
        yield
    except Exception:
        rewrite_tb(*sys.exc_info(), resolution_stack)


def rewrite_traceback_stack(
    exc_type, exc_value, tb, resolution_stack=None, source=None
):
    """Rewrite the current exception to replace any tracebacks from
    within compiled template code with tracebacks that look like they
    came from the template source.

    This must be called within an ``except`` block.

    :param source: For ``TemplateSyntaxError``, the original source if
        known.
    :return: The original exception with the rewritten traceback.
    """
    # Skip the frame for the render function.
    tb = tb.tb_next

    stack = []

    # Build the stack of traceback object, replacing any in template
    # code with the source file and line information.
    while tb is not None:
        # Skip frames decorated with @internalcode. These are internal
        # calls that aren't useful in template debugging output.
        if tb.tb_frame.f_code in INTERNAL_CODE:
            tb = tb.tb_next
            continue

        if tb.tb_frame.f_code in RESOLUTION_START and resolution_stack is not None:
            for snapshot, objs in resolution_stack.get_object_stack():
                fake_tb = fake_traceback(exc_value, tb, snapshot, objs)
                stack.append(fake_tb)

        stack.append(tb)

        tb = tb.tb_next

    tb_next = None

    # Assign tb_next in reverse to avoid circular references.
    for tb in reversed(stack):
        tb_next = tb_set_next(tb, tb_next)

    return exc_value.with_traceback(tb_next)


def fake_traceback(exc_value, tb, snapshot, objs):
    """Produce a new traceback object that looks like it came from the
    template source instead of the compiled code. The filename, line
    number, and location name will point to the template, and the local
    variables will be the current template context.

    :param exc_value: The original exception to be re-raised to create
        the new traceback.
    :param tb: The original traceback to get the local variables and
        code info from.
    :param filename: The template filename.
    :param lineno: The line number in the template source.
    """
    # if tb is not None:
    #     # Replace the real locals with the context that would be
    #     # available at that point in the template.
    #     locals = get_template_locals(tb.tb_frame.f_locals)
    #     locals.pop("__statey_exception__", None)
    # else:
    locals = {}
    lineno = snapshot.lineno

    globals = {
        "__name__": snapshot.filename,
        "__file__": snapshot.filename,
        "__statey_exception__": exc_value,
    }
    # Raise an exception at the correct line number.
    code = compile(
        "\n" * (lineno - 1) + "raise __statey_exception__", snapshot.filename, "exec"
    )

    # Build a new code object that points to the template file and
    # replaces the location with a block name.
    header = "  => "
    whitespace = header
    assert len(header) == len(whitespace)
    obj_lines = "\n".join(
        (header if idx == 0 else whitespace) + repr(obj) for idx, obj in enumerate(objs)
    )
    location = snapshot.name + "\n" + tw.indent(obj_lines, "  ")

    try:
        # Collect arguments for the new code object. CodeType only
        # accepts positional arguments, and arguments were inserted in
        # new Python versions.
        code_args = []

        for attr in (
            "argcount",
            "posonlyargcount",  # Python 3.8
            "kwonlyargcount",
            "nlocals",
            "stacksize",
            "flags",
            "code",  # codestring
            "consts",  # constants
            "names",
            "varnames",
            ("filename", snapshot.filename),
            ("name", location),
            "firstlineno",
            "lnotab",
            "freevars",
            "cellvars",
        ):
            if isinstance(attr, tuple):
                # Replace with given value.
                code_args.append(attr[1])
                continue

            try:
                # Copy original value if it exists.
                code_args.append(getattr(code, "co_" + attr))
            except AttributeError:
                # Some arguments were added later.
                continue

        code = CodeType(*code_args)
    except Exception:
        # Some environments such as Google App Engine don't support
        # modifying code objects.
        pass

    # Execute the new code, which is guaranteed to raise, and return
    # the new traceback without this frame.
    try:
        exec(code, globals, locals)
    except BaseException:
        return sys.exc_info()[2].tb_next


if sys.version_info >= (3, 7):
    # tb_next is directly assignable as of Python 3.7
    def tb_set_next(tb, tb_next):
        tb.tb_next = tb_next
        return tb


elif platform.python_implementation() == "PyPy":
    # PyPy might have special support, and won't work with ctypes.
    try:
        import tputil
    except ImportError:
        # Without tproxy support, use the original traceback.
        def tb_set_next(tb, tb_next):
            return tb

    else:
        # With tproxy support, create a proxy around the traceback that
        # returns the new tb_next.
        def tb_set_next(tb, tb_next):
            def controller(op):
                if op.opname == "__getattribute__" and op.args[0] == "tb_next":
                    return tb_next

                return op.delegate()

            return tputil.make_proxy(controller, obj=tb)


else:
    # Use ctypes to assign tb_next at the C level since it's read-only
    # from Python.
    import ctypes

    class _CTraceback(ctypes.Structure):
        _fields_ = [
            # Extra PyObject slots when compiled with Py_TRACE_REFS.
            ("PyObject_HEAD", ctypes.c_byte * object().__sizeof__()),
            # Only care about tb_next as an object, not a traceback.
            ("tb_next", ctypes.py_object),
        ]

    def tb_set_next(tb, tb_next):
        c_tb = _CTraceback.from_address(id(tb))

        # Clear out the old tb_next.
        if tb.tb_next is not None:
            c_tb_next = ctypes.py_object(tb.tb_next)
            c_tb.tb_next = ctypes.py_object()
            ctypes.pythonapi.Py_DecRef(c_tb_next)

        # Assign the new tb_next.
        if tb_next is not None:
            c_tb_next = ctypes.py_object(tb_next)
            ctypes.pythonapi.Py_IncRef(c_tb_next)
            c_tb.tb_next = c_tb_next

        return tb



import collections
import dataclasses as dc
import inspect
import sys
from contextlib import contextmanager
from functools import wraps, partial, reduce
from itertools import product
from typing import Type, Optional, Union, Any, Dict, Hashable, Sequence, Callable

import marshmallow as ma
import networkx as nx

from statey import NS, exc


LRU_MAX_SIZE = 100

STRING_TYPES = (str, bytes)


@dc.dataclass(frozen=True)
class Global:
    """
    An arbitrary named value intended to be used at the global level
    """

    name: str

    def __repr__(self) -> str:
        return self.name


MISSING = Global("MISSING")


EXPLODE = Global("EXPLODE")


def extract_optional_annotation(annotation: Any) -> Any:
    """
    Determine if the given annotation is an Optional field
    """
    if (
        hasattr(annotation, "__origin__")
        and annotation.__origin__ is Union
        and getattr(annotation, "__args__", None) is not None
        and len(annotation.__args__) == 2
        and annotation.__args__[-1] is type(None)
    ):
        return extract_inner_annotation(annotation)
    return None


def extract_inner_annotation(annotation: Any) -> Any:
    """
    Extract a single inner annotation from an outer annotation
    """
    res = extract_inner_annotations(annotation)
    if not res:
        return None
    return res[0]


def extract_inner_annotations(annotation: Any) -> Any:
    """
    Extract a all inner annotations from `annotation`
    """
    if not getattr(annotation, "__args__", None):
        return None
    return annotation.__args__


def is_sequence_annotation(annotation: Any) -> bool:
    """
    Indicates if the given annotation might be a Sequence[type]
    """
    if not hasattr(annotation, "mro"):
        return False
    if not callable(annotation.mro):
        return False
    if collections.abc.Sequence not in annotation.mro():
        return False
    return True


def is_mapping_annotation(annotation: Any) -> bool:
    """
    Indicates if the given annotation might be a Dict[type] or
    Mapping[type] annotation
    """
    if not hasattr(annotation, "mro"):
        return False
    if not callable(annotation.mro):
        return False
    if not (dict in annotation.mro() or collections.abc.Mapping in annotation.mro()):
        return False
    return True


def dict_get_path(
    data: Dict[Hashable, Any], path: Sequence[Hashable], explode: bool = True
) -> Any:
    """
    Get the key at the given path from the dictionary
    """
    current = data
    current_path = list(path)
    while current_path:
        comp = current_path.pop(0)
        if comp is EXPLODE and explode:
            return [dict_get_path(item, current_path) for item in current]
        elif comp is EXPLODE:
            continue
        current = current[comp]

    return current


def map_dims(
    func: Callable[[Any], Any], arr: Sequence[Any], ndims: int = 1
) -> Sequence[Any]:
    """
    Apply func to the elements of a multi-dimensional array with `ndims` dimensions
    """
    if ndims == 0:
        return func(arr)
    return [map_dims(func, item, ndims - 1) for item in arr]


def dict_set_path(
    data: Dict[Hashable, Any], path: Sequence[Hashable], value: Any
) -> Any:
    """
    Set the given value at the given path in the dict
    """
    current = data
    current_path = path[:-1]
    while current_path:
        comp = current_path.pop(0)
        if comp is EXPLODE:
            for item in current:
                dict_set_path(item, current_path, value)
            return
        current = current.setdefault(comp, {})

    current[path[-1]] = value


@contextmanager
def reraise_ma_validation_error():
    try:
        yield
    except ma.ValidationError as err:
        raise exc.InputValidationError(err.messages) from err


class PossiblySymbolicField(ma.fields.Field):
    """
    A field that will validate the type if given a symbol or the value if given anything else
    """

    def __init__(
        self,
        field: ma.fields.Field,
        type: "Type",
        registry: "Registry",
        *args,
        **kwargs,
    ) -> None:
        super().__init__(*args, **kwargs)
        self.field = field
        self.type = type
        self.registry = registry

    def _serialize(self, value: Any, attr: str, obj: str, **kwargs) -> Any:
        from statey.syms import Object

        # Just pass this through when serializing
        if isinstance(value, Object):
            return value
        self.field.name = self.name
        return self.field._serialize(value, attr, obj, **kwargs)

    def _deserialize(
        self,
        value: Any,
        attr: Optional[str] = None,
        data: Optional[Any] = None,
        **kwargs,
    ) -> Any:
        from statey.syms import types, Object

        self.field.name = self.name
        if not isinstance(value, Object):
            return self.field._deserialize(value, attr, data, **kwargs)

        if self.type != value._type:
            try:
                caster = self.registry.get_caster(value._type, self.type)
            except exc.NoCasterFound as err:
                raise ma.ValidationError(
                    f"Invalid symbol type (expected {self.type}, got {value._type})."
                ) from err
            return caster.cast(value)
        return value

    def __repr__(self) -> str:
        return f"{type(self).__name__}[{self.type}]({type(self.field).__name__}(...))"


class SingleValueFunction(ma.fields.Function):
    """
    Use encoding and decoding functions to process a specific field
    """

    _CHECK_ATTRIBUTE = True

    def _serialize(self, value, attr, obj, **kwargs):
        return self._call_or_raise(self.serialize_func, value, attr)


def filter_dict(
    keep: Callable[[Any], bool], data: Dict[Hashable, Any], and_sequences: bool = True
) -> Dict[Hashable, Any]:
    """
    Similar to the filter() function, but applied recursively to a dict
    """
    out = {}
    for key, val in data.items():
        if isinstance(val, dict):
            filtered = filter_dict(keep, val, and_sequences)
            if filtered:
                out[key] = filtered
        if and_sequences and isinstance(val, Sequence) and and_sequences:
            items = out[key] = []
            for item in val:
                if isinstance(item, dict):
                    item = filter_dict(keep, item, and_sequences)
                    if item:
                        items.append(item)
                elif keep(item):
                    items.append(item)
            if items:
                out[key] = items
        elif keep(val):
            out[key] = val
    return out


def invert_filter(func: Callable[[Any], bool]) -> Callable[[Any], bool]:
    """
    Return a new function that returns `not func(x)`
    """

    def wrapper(x):
        return not func(x)

    return wrapper


def is_missing(value: Any) -> bool:
    """
    Check if a value is equal to MISSING
    """
    return isinstance(value, Global) and value == MISSING


def infer_annotation(obj: Any) -> Any:
    """
    Attempt to infer an annotation from obj, falling back on `type(obj)`
    """
    obj_type = type(obj)
    if (
        isinstance(obj, Sequence)
        and not isinstance(obj, STRING_TYPES)
        and obj
        and reduce(lambda x, y: x is y, map(type, obj))
    ):
        return Sequence[infer_annotation(obj[0])]
    return obj_type


def function_type(
    sig: inspect.Signature,
    return_type: "Type" = MISSING,
    registry: "Registry" = MISSING,
) -> "FunctionType":
    """
    Convert a python function signature to a FunctionType object
    """
    from statey.syms import types

    if is_missing(registry):
        import statey

        registry = statey.registry

    def get_type(annotation):
        if annotation is inspect._empty:
            return types.AnyType()
        return registry.get_type(annotation)

    out_fields = []
    for name, param in sig.parameters.items():
        out_fields.append(types.Field(name, get_type(param.annotation)))

    if is_missing(return_type):
        return_type = get_type(sig.return_annotation)
    return types.NativeFunctionType(tuple(out_fields), return_type)


def single_arg_function_type(
    from_type: "Type", to_type: "Type" = MISSING, arg_name: str = "x", **kwargs
) -> "FunctionType":
    """
    Return a NativeFunctionType for a simple single-argument function
    """
    from statey.syms import types

    if is_missing(to_type):
        to_type = from_type
    fields = (types.Field(arg_name, from_type),)
    return types.NativeFunctionType(fields, to_type, **kwargs)


def native_function(
    input: Callable[[Any], Any], type: "Type" = MISSING, registry: "Registry" = MISSING
) -> "Function":
    """
    Construct a Function object for a regular python function
    """
    from statey.syms import func

    if is_missing(type):
        type = function_type(inspect.signature(input), registry=registry)
    kws = {}
    if hasattr(input, "__name__"):
        kws["name"] = input.__name__
    return func.NativeFunction(type, input, **kws)


def wrap_function_call(
    func: Union["Function", Callable[[Any], Any]],
    args: Sequence[Any] = (),
    kwargs: Optional[Dict[str, Any]] = None,
    registry: Optional["Registry"] = None,
    return_annotation: Any = MISSING,
    frame_depth: int = 0,
) -> "Object":

    from statey.syms import api, impl, types, func as func_module, Object, stack

    if registry is None:
        from statey import registry

    if kwargs is None:
        kwargs = {}

    args = tuple(map(lambda x: Object(x, registry=registry), args))
    kwargs = {key: Object(val, registry=registry) for key, val in kwargs.items()}

    if isinstance(func, func_module.Function):
        function_obj = func
    else:
        try:
            function_obj = native_function(func, registry=registry)
        except ValueError:
            return_annotation = (
                func.return_annotation
                if is_missing(return_annotation)
                else return_annotation
            )
            return_type = registry.get_type(return_annotation)

            fields = []
            for idx, arg in enumerate(args):
                fields.append(types.Field(str(idx), arg._type))

            for name, val in kwargs.items():
                fields.append(types.Field(name, val._type))

            func_type = types.NativeFunctionType(tuple(fields), return_type)
            function_obj = func_module.NativeFunction(func_type, func)

    arguments = bind_function_args(function_obj.type, args, kwargs)
    args_encoder = registry.get_encoder(function_obj.type.args_type)
    encoded_arguments = args_encoder.encode(arguments)

    return Object(
        impl.FunctionCall(function_obj, encoded_arguments),
        registry=registry,
        frame=stack.frame_snapshot(1 + frame_depth),
    )


def bind_function_args(
    func_type: "FunctionType", args: Sequence[Any], kwargs: Dict[str, Any]
) -> Dict[str, Any]:
    """
    Given a function type, bind the given args and kwargs to names
    """
    out_args = {}

    for field, arg in zip(func_type.args, args):
        out_args[field.name] = arg

    names = {arg.name for arg in func_type.args}
    out_args.update({key: val for key, val in kwargs.items() if key in names})
    return out_args


def encodeable_dataclass_fields(data: Any) -> Sequence[dc.field]:
    """.
    Respect the statey-specific field metadtaa and don't encode fields where statey.encode = False
    """
    out = []
    for field in dc.fields(data):
        syms_meta = field.metadata.get(NS) or {}
        encode = syms_meta.get("encode", True)
        if encode:
            out.append(field)
    return out


class Cloneable:
    """
    Mixin for dataclasses adding a clone() method
    """

    def clone(self, **kwargs) -> "Cloneable":
        """
        Return a copy of the current object with the given attributes replaced. This current
        object should be an instance of a dataclass
        """
        return dc.replace(self, **kwargs)


def subgraph_retaining_dependencies(dag: nx.DiGraph, keep_nodes: Sequence[str]) -> None:
    """
    Remove nodes while retaining any indirect dependencies between them. Modifies the graph
    in place.
    """
    for node in list(nx.topological_sort(dag)):
        if node in keep_nodes:
            continue
        for predecessor, successor in product(dag.pred[node], dag.succ[node]):
            dag.add_edge(predecessor, successor)
        dag.remove_node(node)


@dc.dataclass(frozen=True)
class Location(Cloneable):
    """
    Object that simply keeps track of where it is upon successive __getitem__
    operations
    """

    path: Sequence[Any] = ()

    def __getitem__(self, attr: Any) -> "Location":
        return self.clone(path=tuple(self.path) + (attr,))


def trace_func(func: Callable[[Any], Any], cb: Callable[[Any, Any, Any], None]) -> None:
    @wraps(func)
    def wrapper(*args, **kwargs):
        prev = sys.gettrace()
        trace_frame = sys._getframe(0)
        call_frame = None

        def tracer(frame, event, arg):
            nonlocal call_frame
            # Only immediate child frame should be the function call
            if frame.f_back is not trace_frame:
                return
            call_frame = frame
            cb(frame, event, arg)
            # 'done' is a fake event that we fire just to finalize
            # the namespace, so we don't want to pass it to the previous
            # tracer function
            if prev is not None and event != "done":
                prev(frame, event, arg)
            return tracer

        sys.settrace(tracer)
        try:
            return func(*args, **kwargs)
        finally:
            sys.settrace(prev)
            if call_frame is not None:
                tracer(call_frame, "done", None)

    return wrapper


def locals_history(frame):
    """
    Yield a function that, each time it is called, will return the "updated"
    members of the scope of the given frame vs. the previous call
    """
    current_locals = frame.f_locals.copy()

    def diff():
        nonlocal current_locals

        new_locals = frame.f_locals.copy()

        updated = {
            key: loc
            for key, loc in new_locals.items()
            if key not in current_locals or loc is not current_locals[key]
        }
        deleted = {key for key in current_locals if key not in new_locals}
        current_locals = new_locals

        return updated, deleted

    diff.frame = frame

    return diff


def make_tracer(diff_cb: Callable[[Any, Any, Any], None]):
    """
    Wrap the given handler function as a tracing function
    """
    current_history = None

    def tracer(frame, event, arg):
        nonlocal current_history

        if current_history is None:
            current_history = locals_history(frame)

        diff_cb(frame, *current_history())

        return tracer

    return tracer


def locals_diff_tracer(diff_cb: Callable[[Any, Any, Any], None]):
    """
    Decorator to wrap the given function as a trace function
    """
    tracer = make_tracer(diff_cb)

    def dec(func):
        return trace_func(func, cb=tracer)

    return dec


def scope_update_handler(handler: Callable[[Any, str, Any], None]):
    """
    Handle updates to the locals scope using a tracing function
    """

    @locals_diff_tracer
    def differ(frame, updated, deleted):
        for key, val in updated.items():
            handler(frame, key, val)

    return differ


def check_dag(graph: nx.DiGraph) -> None:
    """
    Check if the given graph is a DAG, raising NotADagError otherwise
    """
    try:
        cycle = nx.find_cycle(graph)
    except nx.NetworkXNoCycle:
        return

    edges_left, edges_right = zip(*cycle)
    cycle_path = edges_left + edges_right[-1:]

    raise exc.NotADagError(graph, cycle_path)



import abc
import dataclasses as dc
import json
import os
from typing import Optional

from statey.hooks import hookimpl
from statey.resource import ResourceGraph


class StateManager(abc.ABC):
    """
    A state manager handles reading and writing states to storage
    """

    @abc.abstractmethod
    def load(self, registry: "Registry") -> ResourceGraph:
        """
        Load the resource graph from some storage
        """
        raise NotImplementedError

    @abc.abstractmethod
    def dump(self, graph: ResourceGraph, registry: "Registry") -> None:
        """
        Store the resource graph
        """
        raise NotImplementedError


@dc.dataclass(frozen=True)
class FileStateManager(StateManager):
    """
    Simple state manager that just read and writes states to a file
    """

    path: str

    def load(self, registry: "Registry") -> ResourceGraph:
        if not os.path.exists(self.path):
            return ResourceGraph()
        with open(self.path) as f:
            graph_dict = json.load(f)
        return ResourceGraph.from_dict(graph_dict, registry)

    def dump(self, graph: ResourceGraph, registry: "Registry") -> None:
        graph_as_dict = graph.to_dict(registry)
        with open(self.path, "w+") as f:
            json.dump(graph_as_dict, f, indent=2, sort_keys=True)


class DefaultStateManagerPlugin:
    """
    Plugin to fetch a .statey-state.json file state as the default behavior
    """

    @hookimpl
    def get_state_manager(registry: "Registry") -> StateManager:
        return FileStateManager(".statey-state.json")


DEFAULT_PLUGINS = [DefaultStateManagerPlugin]


def register(registry: Optional["Registry"] = None) -> None:
    """
    Register all plugins in this module
    """
    if registry is None:
        from statey import registry

    for plugin in DEFAULT_PLUGINS:
        registry.register(plugin)



import abc
import dataclasses as dc
from functools import lru_cache
from typing import Dict, Any, Optional, Callable, Sequence

import pluggy

from statey import exc, NS
from statey.hooks import hookspec, create_plugin_manager, register_default_plugins
from statey.syms import types, utils


class Registry(abc.ABC):
    """
    A type registry is used to parse annotations into types
    """

    @abc.abstractmethod
    def get_type(
        self, annotation: Any, meta: Optional[Dict[str, Any]] = None
    ) -> types.Type:
        """
        Parse the given annotation and return a Type. This will properly handle dataclasses
        similarly to how case classes are handled in spark encoding
        """
        raise NotImplementedError

    @abc.abstractmethod
    def infer_type(self, obj: Any) -> types.Type:
        """
        Attempt to infer the type of `obj`, falling back on Any
        """
        raise NotImplementedError

    @abc.abstractmethod
    def get_encoder(self, type: types.Type, serializable: bool = False) -> "Encoder":
        """
        Given a type, return an Encoder instance to encode the type, raising an exc.NoEncoderFound to
        indicate failure
        """
        raise NotImplementedError

    @abc.abstractmethod
    def get_semantics(self, type: types.Type) -> "Semantics":
        """
        Given a type, get the semantics to use for objects of that type.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def get_type_serializer(self, type: types.Type) -> "TypeSerializer":
        """
        Given a type, get the type serializer to turn it into a JSON-serializable value.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def get_type_serializer_from_data(self, data: Any) -> "TypeSerializer":
        """
        Given serialized data, get a type serializer that can decode the serialized type of
        into a native Type
        """
        raise NotImplementedError

    @abc.abstractmethod
    def get_differ(self, type: types.Type) -> "Differ":
        """
        Given a type, get a differ instance to diff the data
        """
        raise NotImplementedError

    #   @abc.abstractmethod
    #   def get_resource(self, name: str) -> "Resource":
    #       """
    # Get the resource with the given name
    # """
    #       raise NotImplementedError

    @abc.abstractmethod
    def get_methods(self, type: types.Type) -> "ObjectMethods":
        """
        Get ObjectMethods for the given type
        """
        raise NotImplementedError

    @abc.abstractmethod
    def get_object(self, value: Any) -> "Object":
        """
        Get a statey object given a value
        """
        raise NotImplementedError

    @abc.abstractmethod
    def get_caster(self, from_type: types.Type, to_type: types.Type) -> "Object":
        """
        Get a caster given from_type and to_type
        """
        raise NotImplementedError

    @abc.abstractmethod
    def get_impl_serializer(
        self, impl: "ObjectImplementation", type: types.Type
    ) -> "ObjectImplementationSerializer":
        """
        Get an object implementation serializer for the given implementation
        and type
        """
        raise NotImplementedError

    @abc.abstractmethod
    def get_impl_serializer_from_data(
        self, data: Any
    ) -> "ObjectImplementationSerializer":
        """
        Get an object implementation serializer from the given data
        """
        raise NotImplementedError

    @abc.abstractmethod
    def get_object_serializer(self, obj: "Object") -> "ObjectSerializer":
        """
        Get an object serializer for the given object
        """
        raise NotImplementedError

    @abc.abstractmethod
    def get_object_serializer_from_data(self, data: Any) -> "ObjectSerializer":
        """
        Get an object serializer from the given data
        """
        raise NotImplementedError

    @abc.abstractmethod
    def get_session_serializer(self, session: "Session") -> "SessionSerializer":
        """
        Get a session serializer for the given session
        """
        raise NotImplementedError

    @abc.abstractmethod
    def get_session_serializer_from_data(self, data: Any) -> "SessionSerializer":
        """
        Get a session serializer from the given data
        """
        raise NotImplementedError

    @abc.abstractmethod
    def get_namespace_serializer(self, ns: "Namespace") -> "NamespaceSerializer":
        """
        Get a namespace serializer for the given namespace
        """
        raise NotImplementedError

    @abc.abstractmethod
    def get_namespace_serializer_from_data(self, data: Any) -> "NamespaceSerializer":
        """
        Get a namespace serializer from the given data
        """
        raise NotImplementedError

    @abc.abstractmethod
    def get_provider(
        self, name: str, params: Optional[Dict[str, Any]] = None
    ) -> "Provider":
        """
        Get a provider with the given name and params
        """
        raise NotImplementedError

    @abc.abstractmethod
    def get_state_manager(self) -> "StateManager":
        """
        Get a state manager to use for the current module
        """
        raise NotImplementedError

    @abc.abstractmethod
    def get_plugin_installer(self, name: str) -> "PluginInstaller":
        """
        Get a plugin installer for a given plugin name
        """
        raise NotImplementedError

    @abc.abstractmethod
    def register(self, plugin: Any) -> None:
        """
        Register a plugin within this registry
        """
        raise NotImplementedError

    def object(self, *args, **kwargs) -> "Object":
        """
        Create an object bound to this registry
        """
        from statey.syms import Object

        return Object(*args, registry=self, **kwargs)


class RegistryHooks:
    """
    Specifies hooks for handling different annotations and converting them to types
    """

    @hookspec(firstresult=True)
    def get_type(
        self, annotation: Any, registry: Registry, meta: Dict[str, Any]
    ) -> types.Type:
        """
        Handle the given annotation and return a Type, or None to indicate it can't be handled by this hook
        """

    @hookspec(firstresult=True)
    def infer_type(self, obj: Any, registry: Registry) -> types.Type:
        """
        Infer a type given an object
        """

    @hookspec(firstresult=True)
    def get_encoder(
        self, type: types.Type, registry: Registry, serializable: bool
    ) -> "Encoder":
        """
                Handle the given type and produce an Encoder instance that can encode values of that type
        `serializable` indicates that encoded values that the Encoder produces should
        always be JSON-serializable
        """

    @hookspec(firstresult=True)
    def get_semantics(self, type: types.Type, registry: Registry) -> "Semantics":
        """
        Handle the given type and produce a Semantics instance for objects of that type
        """

    @hookspec(firstresult=True)
    def get_type_serializer(
        self, type: types.Type, registry: Registry
    ) -> "TypeSerializer":
        """
        Handle the given type and produce an TypeSerializer instance that can encode values of that type
        """

    @hookspec(firstresult=True)
    def get_type_serializer_from_data(
        self, data: Any, registry: Registry
    ) -> "TypeSerializer":
        """
        Handle the given type and produce a TypeSerializer instance for objects of that type
        """

    @hookspec(firstresult=True)
    def get_differ(self, type: types.Type, registry: Registry) -> "Encoder":
        """
        Handle the given type and produce an Differ instance for diffing values of that type
        """

    @hookspec
    def get_methods(self, type: types.Type, registry: Registry) -> "ObjectMethods":
        """
        Hook to get methods for a particular type
        """

    @hookspec(firstresult=True)
    def get_object(self, value: Any, registry: Registry) -> "Object":
        """
        Hook to create an objecxrt from an arbitrary value
        """

    @hookspec(firstresult=True)
    def get_caster(
        self, from_type: types.Type, to_type: types.Type, registry: Registry
    ) -> "Caster":
        """
        Hook to get a caster object to cast from one type to another
        """

    @hookspec(firstresult=True)
    def get_impl_serializer(
        self, impl: "ObjectImplementation", type: types.Type, registry: Registry
    ) -> "ObjectImplementationSerializer":
        """
        Hook to get an implementation serializer from an implementation and type
        """

    @hookspec(firstresult=True)
    def get_impl_serializer_from_data(
        self, data: Any, registry: Registry
    ) -> "ObjectImplementationSerializer":
        """
        Hook to get an impl serializer from data
        """

    @hookspec(firstresult=True)
    def get_object_serializer(
        self, obj: "Object", registry: Registry
    ) -> "ObjectSerializer":
        """
        Hook to get an object serializer from an object
        """

    @hookspec(firstresult=True)
    def get_object_serializer_from_data(
        self, data: Any, registry: Registry
    ) -> "ObjectSerializer":
        """
        Hook to get an object serializer from data
        """

    @hookspec(firstresult=True)
    def get_session_serializer(
        self, session: "Session", registry: Registry
    ) -> "SessionSerializer":
        """
        Hook to get a session serializer from a session
        """

    @hookspec(firstresult=True)
    def get_session_serializer_from_data(
        self, data: Any, registry: Registry
    ) -> "SessionSerializer":
        """
        Hook to get a session serializer from data
        """

    @hookspec(firstresult=True)
    def get_namespace_serializer(
        self, ns: "Namespace", registry: Registry
    ) -> "NamespaceSerializer":
        """
        Hook to get a namespace serializer from a namespace
        """

    @hookspec(firstresult=True)
    def get_namespace_serializer_from_data(
        self, data: Any, registry: Registry
    ) -> "NamespaceSerializer":
        """
        Hook to get a namespace serializer from data
        """

    @hookspec(firstresult=True)
    def get_provider(
        self, name: str, params: Dict[str, Any], registry: Registry
    ) -> "Provider":
        """
        Hook to get a provider from a name and params
        """

    @hookspec(firstresult=True)
    def get_provider_config(
        self, name: str, params: Dict[str, Any], registry: Registry
    ) -> Dict[str, Any]:
        """
        Hook to modify or set a default provider configuration
        """

    @hookspec(firstresult=True)
    def get_state_manager(self, registry: Registry) -> "StateManager":
        """
        Hook to fetch a state manager to use for planning operations
        """

    @hookspec(firstresult=True)
    def get_plugin_installer(self, name: str, registry: Registry) -> "PluginInstaller":
        """
        Hook to fetch a plugin installer
        """

    @hookspec(historic=True)
    def register(self, plugin: Any, registry: Registry) -> None:
        """
        Hook called before a plugin is registeed to add additional side effects.
        """


def create_registry_plugin_manager():
    """
    Factory function to create the default base plugin manager for DefaultTypeRegistry
    """
    pm = create_plugin_manager()
    pm.add_hookspecs(RegistryHooks)
    return pm


@dc.dataclass(frozen=True)
class HookBasedRegistry(Registry):
    """
    Registry based on pluggy hooks
    """

    pm: pluggy.PluginManager = dc.field(
        init=False,
        default_factory=create_registry_plugin_manager,
        compare=False,
        repr=False,
    )

    def register(self, plugin: Any) -> None:
        self.pm.register(plugin)
        self.pm.hook.register.call_historic(kwargs=dict(plugin=plugin, registry=self))

    def load_setuptools_entrypoints(self) -> None:
        """
        Automatically register plugins via setuptools entry points
        """
        self.pm.load_setuptools_entrypoints(NS)

    def get_type(
        self, annotation: Any, meta: Optional[Dict[str, Any]] = None
    ) -> types.Type:
        if meta is None:
            meta = {}
        handled = self.pm.hook.get_type(annotation=annotation, meta=meta, registry=self)
        if handled is None:
            raise exc.NoTypeFound(annotation)
        return handled

    def infer_type(self, obj: Any) -> types.Type:
        handled = self.pm.hook.infer_type(obj=obj, registry=self)
        if handled is not None:
            return handled
        annotation = utils.infer_annotation(obj)
        return self.get_type(annotation)

    @lru_cache(maxsize=1000)
    def get_encoder(self, type: types.Type, serializable: bool = False) -> "Encoder":
        handled = self.pm.hook.get_encoder(
            type=type, registry=self, serializable=serializable
        )
        if handled is None:
            raise exc.NoEncoderFound(type, serializable)
        return handled

    @lru_cache(maxsize=1000)
    def get_semantics(self, type: types.Type) -> "Semantics":
        handled = self.pm.hook.get_semantics(type=type, registry=self)
        if handled is None:
            raise exc.NoSemanticsFound(type)
        return handled

    @lru_cache(maxsize=1000)
    def get_type_serializer(self, type: types.Type) -> "TypeSerializer":
        handled = self.pm.hook.get_type_serializer(type=type, registry=self)
        if handled is None:
            raise exc.NoTypeSerializerFoundForType(type)
        return handled

    def get_type_serializer_from_data(self, data: Any) -> "TypeSerializer":
        handled = self.pm.hook.get_type_serializer_from_data(data=data, registry=self)
        if handled is None:
            raise exc.NoTypeSerializerFoundForData(data)
        return handled

    @lru_cache(maxsize=1000)
    def get_differ(self, type: types.Type) -> "Differ":
        handled = self.pm.hook.get_differ(type=type, registry=self)
        if handled is None:
            raise exc.NoDifferFound(type)
        return handled

    @lru_cache(maxsize=1000)
    def get_differ(self, type: types.Type) -> "Differ":
        handled = self.pm.hook.get_differ(type=type, registry=self)
        if handled is None:
            raise exc.NoDifferFound(type)
        return handled

    @lru_cache(maxsize=1000)
    def get_methods(self, type: types.Type) -> "ObjectMethods":
        from statey.syms.methods import CompositeObjectMethods

        methods_instances = self.pm.hook.get_methods(type=type, registry=self)
        non_null_methods = [
            methods for methods in methods_instances if methods is not None
        ]
        return CompositeObjectMethods(tuple(non_null_methods))

    def get_object(self, value: Any) -> "Object":
        handled = self.pm.hook.get_object(value=value, registry=self)
        if handled is None:
            raise exc.NoObjectFound(value)
        return handled

    def get_impl_serializer(
        self, impl: "ObjectImplementation", type: types.Type
    ) -> "ObjectImplementationSerializer":
        handled = self.pm.hook.get_impl_serializer(impl=impl, type=type, registry=self)
        if handled is None:
            raise exc.NoObjectImplementationSerializerFoundForImpl(impl, type)
        return handled

    @lru_cache(maxsize=1000)
    def get_caster(self, from_type: types.Type, to_type: types.Type) -> "Object":
        handled = self.pm.hook.get_caster(
            from_type=from_type, to_type=to_type, registry=self
        )
        if handled is None:
            raise exc.NoCasterFound(from_type, to_type)
        return handled

    def get_object_serializer(self, obj: "Object") -> "ObjectSerializer":
        handled = self.pm.hook.get_object_serializer(obj=obj, registry=self)
        if handled is None:
            raise exc.NoObjectSerializerFoundForObject(obj)
        return handled

    def get_session_serializer(self, session: "Session") -> "SessionSerializer":
        handled = self.pm.hook.get_session_serializer(session=session, registry=self)
        if handled is None:
            raise exc.NoSessionSerializerFoundForSession(session)
        return handled

    def get_namespace_serializer(self, ns: "Namespace") -> "NamespaceSerializer":
        handled = self.pm.hook.get_namespace_serializer(ns=ns, registry=self)
        if handled is None:
            raise exc.NoNamespaceSerializerFoundForNamespace(ns)
        return handled

    def get_impl_serializer_from_data(
        self, data: Any
    ) -> "ObjectImplementationSerializer":
        handled = self.pm.hook.get_impl_serializer_from_data(data=data, registry=self)
        if handled is None:
            raise exc.NoObjectImplementationSerializerFoundForData(data)
        return handled

    def get_object_serializer_from_data(self, data: Any) -> "ObjectSerializer":
        handled = self.pm.hook.get_object_serializer_from_data(data=data, registry=self)
        if handled is None:
            raise exc.NoObjectSerializerFoundForData(data)
        return handled

    def get_session_serializer_from_data(self, data: Any) -> "SessionSerializer":
        handled = self.pm.hook.get_session_serializer_from_data(
            data=data, registry=self
        )
        if handled is None:
            raise exc.NoSessionSerializerFoundForData(data)
        return handled

    def get_namespace_serializer_from_data(self, data: Any) -> "NamespaceSerializer":
        handled = self.pm.hook.get_namespace_serializer_from_data(
            data=data, registry=self
        )
        if handled is None:
            raise exc.NoNamespaceSerializerFoundForData(data)
        return handled

    def get_provider(
        self, name: str, params: Optional[Dict[str, Any]] = None
    ) -> "Provider":
        if params is None:
            params = {}
        config = self.pm.hook.get_provider_config(
            name=name, params=params, registry=self
        )
        if config is None:
            config = {}
        handled = self.pm.hook.get_provider(name=name, params=config, registry=self)
        if handled is None:
            raise exc.NoProviderFound(name, params)
        return handled

    def get_state_manager(self) -> "StateManager":
        manager = self.pm.hook.get_state_manager(registry=self)
        if manager is None:
            raise exc.NoStateManagerFound()
        return manager

    def get_plugin_installer(self, name: str) -> "PluginInstaller":
        installer = self.pm.hook.get_plugin_installer(name=name, registry=self)
        if installer is None:
            raise exc.NoPluginInstallerFound(name)
        return installer


def create_registry() -> Registry:
    """
    Create a registry using the default implementation
    """
    return HookBasedRegistry()



import abc
import dataclasses as dc
from typing import Any, Optional

import marshmallow as ma

import statey as st
from statey.syms import session, py_session


class NamespaceSerializer(abc.ABC):
    """
    Serializes and deserializes namespaces
    """

    @abc.abstractmethod
    def serialize(self, ns: session.Namespace) -> Any:
        """
        Serialize the given namespace to a JSON-serializble object
        """
        raise NotImplementedError

    @abc.abstractmethod
    def deserialize(self, data: Any) -> session.Namespace:
        """
        Deserialize an object into a namespace
        """
        raise NotImplementedError


class PythonNamespaceSchema(ma.Schema):
    """
    Schema for python namespace dict output
    """

    name = ma.fields.Str(required=True, default="python_namespace")
    types = ma.fields.Dict(keys=ma.fields.Str(), values=ma.fields.Dict())


@dc.dataclass(frozen=True)
class PythonNamespaceSerializer(NamespaceSerializer):
    """
    Serializes/deserializes a python namespace
    """

    registry: "Registry"

    def serialize(self, ns: session.Namespace) -> Any:
        types = {}
        for key, typ in ns.types.items():
            type_ser = self.registry.get_type_serializer(typ)
            types[key] = type_ser.serialize(typ)
        return PythonNamespaceSchema().dump({"types": types})

    def deserialize(self, data: Any) -> session.Namespace:
        loaded_data = PythonNamespaceSchema().load(data)
        types = {}
        ns = py_session.PythonNamespace(self.registry)
        for key, value in loaded_data["types"].items():
            type_ser = self.registry.get_type_serializer_from_data(value)
            typ = type_ser.deserialize(value)
            ns.new(key, typ)
        return ns

    @classmethod
    @st.hookimpl
    def get_namespace_serializer(
        cls, ns: session.Namespace, registry: "Registry"
    ) -> NamespaceSerializer:
        if not isinstance(ns, py_session.PythonNamespace):
            return None
        return cls(registry)

    @classmethod
    @st.hookimpl
    def get_namespace_serializer_from_data(
        cls, data: Any, registry: "Registry"
    ) -> NamespaceSerializer:
        if not isinstance(data, dict) or data.get("name") != "py_session":
            return None
        return cls(registry)


NAMESPACE_SERIALIZER_CLASSES = [PythonNamespaceSerializer]


def register(registry: Optional["Registry"] = None) -> None:
    """
    Register all namespace serializer classes
    """
    if registry is None:
        registry = st.registry

    for cls in NAMESPACE_SERIALIZER_CLASSES:
        registry.register(cls)



# import abc
# import dataclasses as dc
# from typing import Optional, Callable

# from statey.plan import Plan
# from statey.resource import ResourceGraph, ResourceSession


# class ModuleBase(abc.ABC):
#     """

#     """
#     @abc.abstractmethod
#     async def plan(self, state: str, graph: ResourceGraph) -> Plan:
#         """

#         """
#         raise NotImplementedError


# @dc.dataclass(frozen=True)
# class ModuleState:
#     """

#     """
#     name: str
#     factory: Callable[[ResourceSession], None]
#     from_state: Optional[str] = None

#     def __post_init__(self) -> None:
#         if self.from_state is None:
#             self.__dict__['from_state'] = 'DOWN'


# class ModuleMeta(type):
#     """

#     """
#     @classmethod
#     def _validate_states(
#         cls, old_states: Sequence[ModuleState], new_states: Sequence[ModuleState]
#     ) -> Sequence[ModuleState]:

#         all_names = {state.name for state in old_states} | {state.name for state in new_states}
#         for state in new_states:
#             if state.from_state != "DOWN" and state.from_state not in all_names:
#                 raise ValueError(f'Unknown "from_state": {state.from_state}')

#         new_names = Counter((state.name, state.from_state) for state in new_states)
#         state_map = {(state.name, state.from_state): state for state in new_states}
#         if new_names and max(new_names.values()) > 1:
#             multi = [(state_map[k], v) for k, v in new_names.items() if v > 1]
#             raise ValueError(f"Duplicate states found: {multi}")

#         old_states = [state for state in old_states if (state.name, state.from_state) not in new_names]
#         return old_states + list(new_states)

#     def __new__(
#         cls, name: str, bases: Sequence[PyType], attrs: Dict[str, Any]
#     ) -> PyType:
#         super_cls = super().__new__(cls, name, bases, attrs)
#         states = super_cls.__states__ if hasattr(super_cls, "__states__") else ()
#         new_states = [val for val in attrs.values() if isinstance(val, ModuleState)]
#         states = cls._validate_states(states, new_states)
#         super_cls.__states__ = tuple(states)

#         return super_cls


# class Module(ModuleBase, metaclass=ModuleMeta):
#     """

#     """
#     async def plan(self, state: str, graph: ResourceGraph) -> Plan:


# # from statey.fsm import Machine
# # from statey.resource import AbstractState, State


# #     name: str
# #     input_type: types.Type
# #     output_type: types.Type
# #     null: bool


# # # import statey as st


# # # @dc.dataclass(frozen=True)
# # # class ModuleState(abc.ABC):
# # # 	"""
# # # 	Describes a single possible configuration for a single module
# # # 	"""
# # # 	input_type: types.Type
# # # 	output_type: types.Type

# # # 	@abc.abstractmethod
# # # 	def main(self, session: ResourceSession, input_ref: Object) -> Object:
# # # 		"""
# # # 		Populate the given resource session for this state and return a symbol of type
# # # 		self.output_type that will contain the output symbol
# # # 		"""
# # # 		raise NotImplementedError


# # # @dc.dataclass(frozen=True)
# # # class



import abc
import dataclasses as dc
import enum
import types as pytypes
from collections import Counter
from functools import wraps, partial
from typing import Sequence, Callable, Type as PyType, Dict, Any, Optional

import networkx as nx

import statey as st
from statey import resource, task, exc
from statey.provider import Provider
from statey.syms import utils, types, Object, diff


class Transition(abc.ABC):
    """
    A transition defines the procedure from migration a machine
    from one state to another (they may also be the same state)
    """

    from_name: str
    to_name: str
    name: str

    @abc.abstractmethod
    async def plan(
        self,
        current: resource.BoundState,
        config: resource.BoundState,
        session: task.TaskSession,
    ) -> Object:
        """
        Same as Resource.plan(), except for planning
        a specific transition.
        """
        raise NotImplementedError


@dc.dataclass(frozen=True)
class FunctionTransition(Transition):
    """
    Transition class that simply wraps a function
    """

    from_name: str
    to_name: str
    name: str
    func: Callable[[Any], Any]

    async def plan(
        self,
        current: resource.StateSnapshot,
        config: resource.StateConfig,
        session: task.TaskSession,
    ) -> Object:
        return await self.func(current=current, config=config, session=session)


def transition(from_name: str, to_name: str, name: str = utils.MISSING) -> Any:
    """
    Generate a decorate to wrap a function as a transition
    """

    def dec(func):
        nonlocal name
        if name is utils.MISSING:
            name = getattr(func, "__name__", "<unknown>")

        @wraps(func)
        def get_transition(*args, **kwargs):
            new_func = lambda *args2, **kwargs2: func(
                *args, *args2, **kwargs, **kwargs2
            )
            return FunctionTransition(from_name, to_name, name, new_func)

        get_transition.transition_factory = True

        return get_transition

    return dec


class MachineMeta(type(resource.Resource)):
    """
    Special behavior for state machines
    """

    @classmethod
    def _validate_states(
        cls, old_states: Sequence[resource.State], new_states: Sequence[resource.State]
    ) -> Sequence[resource.State]:

        new_names = Counter(state.name for state in new_states)
        if new_names and max(new_names.values()) > 1:
            multi = {k: v for k, v in new_names.items() if v > 1}
            raise ValueError(f"Duplicate states found: {multi}")

        old_states = [state for state in old_states if state.name not in new_names]
        return old_states + list(new_states)

    def __new__(
        cls, name: str, bases: Sequence[PyType], attrs: Dict[str, Any]
    ) -> PyType:
        super_cls = super().__new__(cls, name, bases, attrs)
        states = super_cls.__states__ if hasattr(super_cls, "__states__") else ()
        new_states = [val for val in attrs.values() if isinstance(val, resource.State)]
        states = cls._validate_states(states, new_states)
        super_cls.__states__ = tuple(states)

        transitions = (
            super_cls.__transitions__
            if hasattr(super_cls, "__transitions__")
            else set()
        )
        new_transitions = {
            name
            for name, val in attrs.items()
            if hasattr(val, "transition_factory") and val.transition_factory
        }
        super_cls.__transitions__ = transitions | new_transitions

        return super_cls


class Machine(resource.Resource, metaclass=MachineMeta):
    """
    Class with a metaclass to automatically collect states and transitions into class variables.
    """

    def __init__(self, name: str, provider: Optional[Provider] = None) -> None:
        if provider is None:
            from statey.provider import default_provider as provider

        self.name = name
        self.provider = provider
        # This is temporary, should clean this up
        for state in self.__states__:
            self.set_resource_state(resource.ResourceState(state, name, provider.id))

    def set_resource_state(self, state: resource.ResourceState) -> None:
        setattr(self, state.state.name, state)

    @property
    def null_state(self) -> resource.ResourceState:
        state = next((s for s in self.__states__ if s.null))
        return resource.ResourceState(state, self.name, self.provider.id)

    async def plan(
        self,
        current: resource.StateSnapshot,
        config: resource.StateConfig,
        session: task.TaskSession,
    ) -> Object:

        from_name = current.state.name
        to_name = config.state.name

        transitions = (getattr(self, tran)() for tran in self.__transitions__)
        transition = next(
            (
                tran
                for tran in transitions
                if tran.from_name == from_name
                if tran.to_name == to_name
            ),
            None,
        )
        if transition is None:
            raise exc.PlanError(
                f"Unable to find transition from {from_name} to {to_name}."
            )

        return await transition.plan(current, config, session)

    def __call__(self, *args, **kwargs) -> resource.ResourceState:
        states = [state for state in self.__states__ if state != self.null_state.state]
        if len(states) > 1:
            raise TypeError(f'"{self.name}" has more than one non-null state.')
        if len(states) < 1:
            raise TypeError(f'"{self.name}" does not have any non-null states.')
        return resource.ResourceState(states[0], self.name, self.provider.id)(
            *args, **kwargs
        )

    @abc.abstractmethod
    async def refresh(self, current: resource.BoundState) -> resource.BoundState:
        """
        Same as Resource.refresh()
        """
        raise NotImplementedError

    async def finalize(self, current: resource.BoundState) -> resource.BoundState:
        return current


class ModificationAction(enum.Enum):
    """
    Actions to control simple machine behavior
    """

    NONE = "none"
    MODIFY = "modify"
    DELETE_AND_RECREATE = "delete_and_recreate"


class SingleStateMachine(Machine):
    """
    A simple machine is an FSM which can only have two states: UP and DOWN.

    Note that a SimpleMachine's UP state should have all of the same fields available
    in its output type as its input type.
    """

    UP: resource.State
    DOWN: resource.NullState = resource.NullState("DOWN")

    @abc.abstractmethod
    async def create(
        self, session: task.TaskSession, config: resource.StateConfig
    ) -> "Object":
        """
        Create this resource with the given configuration
        """
        raise NotImplementedError

    @abc.abstractmethod
    async def delete(
        self, session: task.TaskSession, current: resource.StateSnapshot
    ) -> "Object":
        """
        Delete the resource with the given data
        """
        raise NotImplementedError

    @abc.abstractmethod
    async def modify(
        self,
        session: task.TaskSession,
        current: resource.StateSnapshot,
        config: resource.StateConfig,
    ) -> "Object":
        """
        Modify the resource from `data` to the given config. Default implementation
        is always to delete and recreate the resource.

        NOTE: if subclasses do not modify the get_action() implementation they can
        override this with a stub method, as it will never be called. It is defined
        as an abstract to avoid the case where it is omitted accidentally and
        NotImplementedError is raised during the task execution
        """
        raise NotImplementedError

    # Overridding this as an "optional" abstract method
    modify = NotImplemented

    @abc.abstractmethod
    async def refresh_state(self, data: Any) -> Optional[Any]:
        """
        Get a refreshed version of `data` (which is in the state UP). Return None
        to indicate the resource no longer exists.
        """
        raise NotImplementedError

    @abc.abstractmethod
    async def get_action(
        self,
        current: resource.StateSnapshot,
        config: resource.StateConfig,
        session: task.TaskSession,
    ) -> ModificationAction:
        """
        From the current, and config values, determine which modification action should be taken.
        """
        raise NotImplementedError

    async def refresh_config(self, config: "Object") -> "Object":
        """
        Transform a configuration before planning
        """
        return config

    async def refresh(self, current: resource.StateSnapshot) -> resource.StateSnapshot:
        if current.state.name == self.null_state.name:
            return current
        info = await self.refresh_state(current.data)
        if info is None:
            return resource.StateSnapshot({}, self.null_state)
        return resource.StateSnapshot(info, current.state)

    @transition("UP", "UP")
    async def modify_resource(
        self,
        current: resource.StateSnapshot,
        config: resource.StateConfig,
        session: task.TaskSession,
    ) -> Object:

        config = config.clone(obj=await self.refresh_config(config.obj))

        action = await self.get_action(current, config, session)

        if action == ModificationAction.NONE:
            return current.obj

        if action == ModificationAction.MODIFY:
            if self.modify is NotImplemented:
                raise NotImplementedError(
                    f"`modify` has not been defined in {type(self).__name__}."
                )
            return await self.modify(session, current, config)

        if action == ModificationAction.DELETE_AND_RECREATE:
            raise exc.NullRequired

        raise exc.InvalidModificationAction(action)

    @transition("DOWN", "UP")
    async def create_resource(
        self,
        current: resource.StateSnapshot,
        config: resource.StateConfig,
        session: task.TaskSession,
    ) -> Object:

        config = config.clone(obj=await self.refresh_config(config.obj))
        return await self.create(session, config)

    @transition("UP", "DOWN")
    async def delete_resource(
        self,
        current: resource.StateSnapshot,
        config: resource.StateConfig,
        session: task.TaskSession,
    ) -> Object:

        return await self.delete(session, current)

    @transition("DOWN", "DOWN")
    async def noop_down(
        self,
        current: resource.StateSnapshot,
        config: resource.StateConfig,
        session: task.TaskSession,
    ) -> Object:

        return current.obj


class SimpleMachine(SingleStateMachine):
    """
    A simple machine has only a single state and each transition only consists
    of a single task
    """

    async def get_expected(
        self,
        current: resource.StateSnapshot,
        config: resource.StateConfig,
        session: task.TaskSession,
    ) -> Any:
        """
        Get the expected output for the given configuration. Default implementation
        is just passing through config fields and setting the rest as unknown
        """
        output = st.Unknown[config.state.output_type]
        if not current.state.null:
            output = current.obj
        return st.fill(config.obj, config.state.output_type, output)

    # Not defined as abstract methods because subclasses may want to just override
    # the top-level methods instead
    async def create_task(self, config: Any) -> Any:
        """
        Defines a single task called "create" that will create this resource
        """
        raise NotImplementedError

    async def delete_task(self, current: Any) -> Any:
        """
        Defines a single task called "delete" that will delete this resource
        """
        raise NotImplementedError

    async def modify_task(self, diff: diff.Diff, current: Any, config: Any) -> Any:
        """
        Defines a single task called "modify" that will modify this resource
        """
        raise NotImplementedError

    def _get_optional_method(self, name: str) -> Callable[[Any], Any]:
        if getattr(type(self), name) is getattr(SimpleMachine, name):
            raise NotImplementedError(f"{name} has not been defined in this class.")
        return getattr(self, name)

    def get_action_from_diff(self, diff: diff.Diff) -> ModificationAction:
        """
        With the given diff, determine which action must be taken to get to the configured
        state. This is only called when both the current and configured state are UP.

        Overriding this method is optional, by default it will always delete and recreate
        the resource.
        """
        if not diff:
            return ModificationAction.NONE
        return ModificationAction.DELETE_AND_RECREATE

    def get_diff(
        self,
        current: resource.StateSnapshot,
        config: resource.StateConfig,
        session: task.TaskSession,
    ) -> diff.Diff:
        """
        Produce a diff given the current, config and session data
        """
        differ = session.ns.registry.get_differ(config.state.input_type)
        current_as_config = st.filter_struct(current.obj, config.type)
        return differ.diff(current_as_config, config.obj, session)

    async def get_action(
        self,
        current: resource.StateSnapshot,
        config: resource.StateConfig,
        session: task.TaskSession,
    ) -> ModificationAction:
        """
        Split get_action into get_diff and get_action_from_diff
        """
        diff = self.get_diff(current, config, session)
        return self.get_action_from_diff(diff)

    async def create(
        self, session: task.TaskSession, config: resource.StateConfig
    ) -> "Object":
        current = resource.StateSnapshot({}, self.null_state.state)
        expected = await self.get_expected(current, config, session)
        create_task = self._get_optional_method("create_task")
        return session["create"] << (task.new(create_task)(config.obj) >> expected)

    async def delete(
        self, session: task.TaskSession, current: resource.StateSnapshot
    ) -> "Object":
        delete_task = self._get_optional_method("delete_task")
        ref = session["delete"] << task.new(delete_task)(current.obj)
        return st.join(st.Object({}, st.EmptyType, session.ns.registry), ref)

    async def modify(
        self,
        session: task.TaskSession,
        current: resource.StateSnapshot,
        config: resource.StateConfig,
    ) -> "Object":
        expected = await self.get_expected(current, config, session)
        modify_task = self._get_optional_method("modify_task")
        diff = self.get_diff(current, config, session)
        partial_modify = partial(modify_task, diff)
        return session["modify"] << (
            task.new(partial_modify)(current.obj, config.obj) >> expected
        )


# class MachineResource(resource.Resource):
#     """
#     Simple wrapper resource, for state machines all logic is really in the States
#     implementation

#     Example:
#     rs = MachineResource(MyMachine('new_resource'))
#     """

#     # This will be set in the constructor
#     States = None

#     def __init__(
#         self, name: str, machine_cls: PyType[Machine], provider: Provider
#     ) -> None:
#         self.States = self.machine_cls = machine_cls
#         self.name = name
#         self.provider = provider
#         super().__init__()

#     async def plan(
#         self,
#         current: resource.StateSnapshot,
#         config: resource.StateConfig,
#         session: task.TaskSession,
#     ) -> Object:
#         return await self.s.plan(current, config, session)

#     async def refresh(self, current: resource.StateSnapshot) -> resource.StateSnapshot:
#         return await self.s.refresh(current)

#     async def finalize(self, current: resource.StateSnapshot) -> resource.StateSnapshot:
#         return await self.s.finalize(current)



from typing import Optional

import pylumi

import statey as st
from statey.ext.pulumi import exc
from statey.ext.pulumi.constants import PULUMI_NS


class PulumiPluginInstaller(st.PluginInstaller):
    """
    Plugin installer using pulumi's native functionality
    """

    def __init__(self) -> None:
        self.ctx = None

    def setup(self) -> None:
        self.ctx = pylumi.Context()
        self.ctx.setup()

    def teardown(self) -> None:
        self.ctx.teardown()
        self.ctx = None

    def install(self, name: str, version: Optional[str] = None) -> None:
        if not version:
            raise exc.PluginVersionRequired(name)
        _, plugin_name = name.split("/", 1)
        self.ctx.install_plugin("resource", plugin_name, version)

    @st.hookimpl
    def get_plugin_installer(
        self, name: str, registry: st.Registry
    ) -> st.PluginInstaller:
        if not name.startswith(PULUMI_NS + "/"):
            return None
        return self


PLUGINS = [PulumiPluginInstaller()]


def register(registry: Optional[st.Registry] = None) -> None:
    """
    Register plugins for this module
    """
    if registry is None:
        registry = st.registry

    for plugin in PLUGINS:
        registry.register(plugin)



from typing import Optional

import pluggy

from statey import NS


hookspec = pluggy.HookspecMarker(NS)

hookimpl = pluggy.HookimplMarker(NS)


def create_plugin_manager() -> pluggy.PluginManager:
    """
    Factory function to create a plugin manager w/ the default namespace
    """
    return pluggy.PluginManager(NS)


def register_default_plugins(
    registry: Optional["Registry"] = None,
    encoders: bool = True,
    type_plugins: bool = True,
    semantics: bool = True,
    type_serializers: bool = True,
    providers: bool = True,
    extensions: bool = True,
    differs: bool = True,
    methods: bool = True,
    casters: bool = True,
    impl_serializers: bool = True,
    object_serializers: bool = True,
    namespace_serializers: bool = True,
    session_serializers: bool = True,
    state_managers: bool = True,
    setuptools_entrypoints: bool = True,
) -> None:
    """
    Convenience method to register all of the default provided hooks for the given
    object types
    """
    if registry is None:
        from statey import registry

    if encoders:
        from statey.syms.encoders import register as register_encoders

        register_encoders(registry)

    if type_plugins:
        from statey.syms.plugins import register as register_type_plugins

        register_type_plugins(registry)

    if semantics:
        from statey.syms.semantics import register as register_semantics

        register_semantics(registry)

    if type_serializers:
        from statey.syms.type_serializers import register as register_serializers

        register_serializers(registry)

    if differs:
        from statey.syms.diff import register as register_differs

        register_differs(registry)

    if casters:
        from statey.syms.casters import register as register_casters

        register_casters(registry)

    if methods:
        from statey.syms.methods import register as register_methods

        register_methods(registry)

    if impl_serializers:
        from statey.syms.impl_serializers import register as register_impl_serializers

        register_impl_serializers(registry)

    if object_serializers:
        from statey.syms.object_serializers import (
            register as register_object_serializers,
        )

        register_object_serializers(registry)

    if namespace_serializers:
        from statey.syms.namespace_serializers import (
            register as register_ns_serializers,
        )

        register_ns_serializers(registry)

    if session_serializers:
        from statey.syms.session_serializers import (
            register as register_session_serializers,
        )

        register_session_serializers(registry)

    if providers:
        from statey.provider import register as register_providers

        register_providers(registry)

    if extensions:
        from statey.ext import register as register_extensions

        register_extensions(registry)

    if state_managers:
        from statey.state_manager import register as register_state_managers

        register_state_managers(registry)

    if setuptools_entrypoints:
        registry.load_setuptools_entrypoints()
